# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
import pytest
from hypothesis import assume, given, note, strategies as st, settings

import montepy

from montepy.data_inputs.element import Element
from montepy.data_inputs.nuclide import Nucleus, Nuclide, Library
from montepy.input_parser import syntax_node
from montepy.errors import *
from montepy.particle import LibraryType


class TestNuclide:
    def test_nuclide_init(_):
        isotope = Nuclide("1001.80c")
        assert isotope.ZAID == 1001
        assert isotope.Z == 1
        assert isotope.A == 1
        assert isotope.element.Z == 1
        assert isotope.library == "80c"
        with pytest.raises(ValueError):
            Nuclide("1001.80c.5")
        with pytest.raises(ValueError):
            Nuclide("hi.80c")

    def test_nuclide_metastable_init(_):
        isotope = Nuclide("13426.02c")
        assert isotope.ZAID == 13426
        assert isotope.Z == 13
        assert isotope.A == 26
        assert isotope.is_metastable
        assert isotope.meta_state == 1
        isotope = Nuclide("92635.02c")
        assert isotope.A == 235
        assert isotope.meta_state == 1
        isotope = Nuclide("92935.02c")
        assert isotope.A == 235
        assert isotope.meta_state == 4
        assert isotope.mcnp_str() == "92935.02c"
        edge_cases = [
            ("4412", 4, 12, 1),
            ("4413", 4, 13, 1),
            ("4414", 4, 14, 1),
            ("36569", 36, 69, 2),
            ("77764", 77, 164, 3),
        ]
        for ZA, Z_ans, A_ans, isomer_ans in edge_cases:
            isotope = Nuclide(ZA + ".80c")
            assert isotope.Z == Z_ans
            assert isotope.A == A_ans
            assert isotope.meta_state == isomer_ans
        with pytest.raises(ValueError):
            isotope = Nuclide("13826.02c")

    def test_nuclide_get_base_zaid(_):
        isotope = Nuclide("92635.02c")
        assert isotope.get_base_zaid() == 92235

    def test_nuclide_library_setter(_):
        isotope = Nuclide("1001.80c")
        isotope.library = "70c"
        assert isotope.library == "70c"
        with pytest.raises(TypeError):
            isotope.library = 1

    def test_nuclide_str(_):
        isotope = Nuclide("1001.80c")
        assert isotope.mcnp_str() == "1001.80c"
        assert isotope.nuclide_str() == "H-1.80c"
        assert repr(isotope) == "Nuclide('H-1.80c')"
        assert str(isotope) == " H-1     (80c)"
        isotope = Nuclide("94239.80c")
        assert isotope.nuclide_str() == "Pu-239.80c"
        assert isotope.mcnp_str() == "94239.80c"
        assert repr(isotope) == "Nuclide('Pu-239.80c')"
        isotope = Nuclide("92635.80c")
        assert isotope.nuclide_str() == "U-235m1.80c"
        assert isotope.mcnp_str() == "92635.80c"
        assert str(isotope) == " U-235m1 (80c)"
        assert repr(isotope) == "Nuclide('U-235m1.80c')"
        # stupid legacy stupidity #486
        isotope = Nuclide("95642")
        assert isotope.nuclide_str() == "Am-242"
        assert isotope.mcnp_str() == "95642"
        assert repr(isotope) == "Nuclide('Am-242')"
        isotope = Nuclide("95242")
        assert isotope.nuclide_str() == "Am-242m1"
        assert isotope.mcnp_str() == "95242"
        assert repr(isotope) == "Nuclide('Am-242m1')"
        # test that can be formatted at all:
        f"{isotope:0>10s}"

    @pytest.mark.parametrize(
        "input, Z, A, meta, library",
        [
            (1001, 1, 1, 0, ""),
            ("1001.80c", 1, 1, 0, "80c"),
            ("h1", 1, 1, 0, ""),
            ("h-1", 1, 1, 0, ""),
            ("h-1.80c", 1, 1, 0, "80c"),
            ("h", 1, 0, 0, ""),
            ("92635m2.710nc", 92, 235, 3, "710nc"),
            (Nuclide("1001.80c"), 1, 1, 0, "80c"),
            (Nucleus(Element(1), 1), 1, 1, 0, ""),
            (Element(1), 1, 0, 0, ""),
            (92, 92, 0, 0, ""),
        ],
    )
    def test_fancy_names(_, input, Z, A, meta, library):
        isotope = Nuclide(input)
        assert isotope.A == A
        assert isotope.Z == Z
        assert isotope.meta_state == meta
        assert isotope.library == Library(library)

    nuclide_strat = (
        st.integers(1, 118),
        st.floats(2.1, 2.7),
        st.integers(0, 4),
        st.integers(0, 999),
        # based on Table B.1 of the 6.3.1 manual
        # ignored `t` because that requires an `MT`
        st.sampled_from(
            [c for c in "cdmgpuyehporsa"]
        ),  # lazy way to avoid so many quotation marks
        st.booleans(),
    )

    @given(*nuclide_strat)
    def test_fancy_names_pbt(
        _, Z, A_multiplier, meta, library_base, library_extension, hyphen
    ):
        # avoid Am-242 metastable legacy
        A = int(Z * A_multiplier)
        element = Element(Z)
        assume(not (Z == 95 and A == 242))
        # ignore H-*m* as it's nonsense
        assume(not (Z == 1 and meta > 0))
        for lim_Z, lim_A in Nuclide._BOUNDING_CURVE:
            if Z <= lim_Z:
                break
        assume(A <= lim_A)
        library = f"{library_base:02}{library_extension}"
        inputs = [
            f"{Z* 1000 + A}{f'm{meta}' if meta > 0 else ''}.{library}",
            f"{Z* 1000 + A}{f'm{meta}' if meta > 0 else ''}",
            f"{element.symbol}{'-' if hyphen else ''}{A}{f'm{meta}' if meta > 0 else ''}.{library}",
            f"{element.symbol}{'-' if hyphen else ''}{A}{f'm{meta}' if meta > 0 else ''}",
        ]

        if meta:
            inputs.append(f"{Z* 1000 + A + 300 + 100 * meta}.{library}")
        note(inputs)
        for input in inputs:
            note(input)
            isotope = Nuclide(input)
            assert isotope.A == A
            assert isotope.Z == Z
            assert isotope.meta_state == meta
            if "." in input:
                assert isotope.library == Library(library)
                new_isotope = Nuclide(Z=Z, A=A, meta_state=meta, library=library)
            else:
                assert isotope.library == Library("")
                new_isotope = Nuclide(Z=Z, A=A, meta_state=meta)
            # test eq and lt
            assert new_isotope == isotope
            new_isotope = Nuclide(Z=Z, A=A + 5, meta_state=meta)
            assert new_isotope != isotope
            assert isotope < new_isotope
            if library_base < 998:
                new_isotope = Nuclide(
                    Z=Z,
                    A=A,
                    meta_state=meta,
                    library=f"{library_base+2:02}{library_extension}",
                )
                assert isotope < new_isotope
            with pytest.raises(TypeError):
                isotope == "str"
            with pytest.raises(TypeError):
                isotope < 5

    @given(*nuclide_strat)
    def test_valuenode_init(
        _, Z, A_multiplier, meta, library_base, library_extension, hyphen
    ):
        # avoid Am-242 metastable legacy
        A = int(Z * A_multiplier)
        element = Element(Z)
        assume(not (Z == 95 and A == 242))
        # ignore H-*m* as it's nonsense
        assume(not (Z == 1 and meta > 0))
        for lim_Z, lim_A in Nuclide._BOUNDING_CURVE:
            if Z <= lim_Z:
                break
        assume(A <= lim_A)
        library = f"{library_base:02}{library_extension}"
        ZAID = Z * 1_000 + A
        if meta > 0:
            ZAID += 300 + meta * 100

        inputs = [
            f"{ZAID}.{library}",
            f"{ZAID}",
        ]
        for input in inputs:
            note(input)
            for type in {float, str}:
                if type == float and "." in input:
                    continue
                node = syntax_node.ValueNode(input, type, syntax_node.PaddingNode(" "))
                nuclide = Nuclide(node=node)
                assert nuclide.Z == Z
                assert nuclide.A == A
                assert nuclide.meta_state == meta
                if "." in input:
                    assert str(nuclide.library) == library
                else:
                    assert str(nuclide.library) == ""

    @pytest.mark.parametrize(
        "kwargs, error",
        [
            ({"name": 1.23}, TypeError),
            ({"name": int(1e6)}, ValueError),
            ({"name": "1001.hi"}, ValueError),
            ({"name": "hello"}, ValueError),
            ({"element": "hi"}, TypeError),
            ({"Z": "hi"}, TypeError),
            ({"Z": 1000}, montepy.errors.UnknownElement),
            ({"Z": 1, "A": "hi"}, TypeError),
            ({"Z": 1, "A": -1}, ValueError),
            ({"A": 1}, ValueError),
            ({"meta_state": 1}, ValueError),
            ({"library": "80c"}, ValueError),
            ({"Z": 1, "A": 2, "meta_state": "hi"}, TypeError),
            ({"Z": 1, "A": 2, "meta_state": -1}, ValueError),
            ({"Z": 1, "A": 2, "meta_state": 5}, ValueError),
            ({"name": "1001", "library": 5}, TypeError),
            ({"name": "1001", "library": "hi"}, ValueError),
        ],
    )
    def test_nuclide_bad_init(_, kwargs, error):
        with pytest.raises(error):
            Nuclide(**kwargs)


class TestLibrary:

    @pytest.mark.parametrize(
        "input, lib_type",
        [
            ("80c", LibraryType.NEUTRON),
            ("710nc", LibraryType.NEUTRON),
            ("50d", LibraryType.NEUTRON),
            ("50M", LibraryType.NEUTRON),
            ("01g", LibraryType.PHOTO_ATOMIC),
            ("84P", LibraryType.PHOTO_ATOMIC),
            ("24u", LibraryType.PHOTO_NUCLEAR),
            ("30Y", LibraryType.NEUTRON),
            ("03e", LibraryType.ELECTRON),
            ("70H", LibraryType.PROTON),
            ("70o", LibraryType.DEUTERON),
            ("70r", LibraryType.TRITON),
            ("70s", LibraryType.HELION),
            ("70a", LibraryType.ALPHA_PARTICLE),
        ],
    )
    def test_library_init(_, input, lib_type):
        lib = Library(input)
        assert lib.library_type == lib_type, "Library type not properly parsed"
        assert str(lib) == input, "Original string not preserved."
        assert lib.library == input, "Original string not preserved."

    def test_library_bad_init(_):
        with pytest.raises(TypeError):
            Library(5)
        with pytest.raises(ValueError):
            Library("hi")
        with pytest.raises(ValueError):
            Library("00x")

    @given(
        input_num=st.integers(min_value=0, max_value=999),
        extra_char=st.characters(min_codepoint=97, max_codepoint=122),
        lib_extend=st.sampled_from("cdmgpuyehorsa"),
        capitalize=st.booleans(),
    )
    def test_library_mass_init(_, input_num, extra_char, lib_extend, capitalize):
        if input_num > 100:
            input = f"{input_num:02d}{extra_char}{lib_extend}"
        else:
            input = f"{input_num:02d}{lib_extend}"
        if capitalize:
            input = input.upper()
        note(input)
        lib = Library(input)
        assert str(lib) == input, "Original string not preserved."
        assert repr(lib) == f"Library('{input}')", "Original string not preserved."
        assert lib.library == input, "Original string not preserved."
        assert lib.number == input_num, "Library number not preserved."
        assert lib.suffix == lib_extend, "Library suffix not preserved."
        lib2 = Library(input)
        assert lib == lib2, "Equality broke."
        assert hash(lib) == hash(lib2), "Hashing broke for library."

    @pytest.mark.parametrize(
        "input, error", [(5, TypeError), ("hi", ValueError), ("75b", ValueError)]
    )
    def test_bad_library_init(_, input, error):
        with pytest.raises(error):
            Library(input)
        lib = Library("00c")
        if not isinstance(input, str):
            with pytest.raises(TypeError):
                lib == input, "Type enforcement for library equality failed."

    def test_library_sorting(_):
        lib = Library("00c")
        with pytest.raises(TypeError):
            lib < 5
        libs = {Library(s) for s in ["00c", "70c", "70g", "80m", "24y", "90a"]}
        libs.add("50d")
        gold_order = ["90a", "00c", "70c", "50d", "70g", "80m", "24y"]
        assert [str(lib) for lib in sorted(libs)] == gold_order, "Sorting failed."

    def test_library_bool(_):
        assert Library("80c")
        assert not Library("")


# test element
class TestElement:
    def test_element_init(_):
        for Z in range(1, 119):
            element = Element(Z)
            assert element.Z == Z
            # Test to ensure there are no missing elements
            name = element.name
            symbol = element.symbol

        with pytest.raises(UnknownElement):
            Element(119)

        spot_check = {
            1: ("H", "hydrogen"),
            40: ("Zr", "zirconium"),
            92: ("U", "uranium"),
            94: ("Pu", "plutonium"),
            29: ("Cu", "copper"),
            13: ("Al", "aluminum"),
        }
        for z, (symbol, name) in spot_check.items():
            element = Element(z)
            assert z == element.Z
            assert symbol == element.symbol
            assert name == element.name

    def test_element_str(_):
        element = Element(1)
        assert str(element) == "hydrogen"
        assert repr(element) == "Element(1)"

    def test_get_by_symbol(_):
        element = Element.get_by_symbol("Hg")
        assert element.name == "mercury"
        with pytest.raises(UnknownElement):
            Element.get_by_symbol("Hi")

    def test_get_by_name(_):
        element = Element.get_by_name("mercury")
        assert element.symbol == "Hg"
        with pytest.raises(UnknownElement):
            Element.get_by_name("hudrogen")

    # particle
    def test_particle_str(_):
        part = montepy.Particle("N")
        assert str(part) == "neutron"


class TestNucleus:

    @given(Z=st.integers(1, 99), A=st.integers(0, 300), meta=st.integers(0, 4))
    def test_nucleus_init_eq_hash(_, Z, A, meta):
        # avoid metastable elemental
        assume((A == 0) == (meta == 0))
        nucleus = Nucleus(Element(Z), A, meta)
        assert nucleus.Z == Z
        assert nucleus.A == A
        assert nucleus.meta_state == meta
        # test eq
        other = Nucleus(Element(Z), A, meta)
        assert nucleus == other
        assert hash(nucleus) == hash(other)
        assert str(nucleus) == str(other)
        assert repr(nucleus) == repr(other)
        with pytest.raises(TypeError):
            nucleus == 5
        with pytest.raises(TypeError):
            nucleus < 5
        # test not eq
        if A != 0:
            new_meta = meta + 1 if meta <= 3 else meta - 1
            for other in {
                Nucleus(Element(Z), A + 5, meta),
                Nucleus(Element(Z), A, new_meta),
            }:
                assert nucleus != other
                assert hash(nucleus) != hash(other)
                assert str(nucleus) != str(other)
                assert repr(nucleus) != repr(other)
                if other.A > A:
                    assert nucleus < other
                else:
                    if new_meta > meta:
                        assert nucleus < other
                    elif new_meta < meta:
                        assert other < nucleus
        # avoid insane ZAIDs
        a_ratio = A / Z
        if a_ratio >= 1.9 and a_ratio < 2.3:
            nuclide = Nuclide(nucleus.ZAID)
            assert nuclide.nucleus == nucleus
            nucleus = Nucleus(Element(Z))
            assert nucleus.Z == Z

    @pytest.mark.parametrize(
        "kwargs, error",
        [
            ({"element": "hi"}, TypeError),
            ({"A": "hi"}, TypeError),
            ({"A": -1}, ValueError),
            ({"meta_state": "hi"}, TypeError),
            ({"meta_state": -1}, ValueError),
            ({"meta_state": 5}, ValueError),
        ],
    )
    def test_nucleus_bad_init(_, kwargs, error):
        if "element" not in kwargs:
            kwargs["element"] = Element(1)
        with pytest.raises(error):
            Nucleus(**kwargs)


class TestLibraryType:

    def test_sort_order(_):
        gold = [
            "alpha_particle",
            "deuteron",
            "electron",
            "proton",
            "neutron",
            "photo_atomic",
            "photo_nuclear",
            "helion",
            "triton",
        ]
        sort_list = sorted(LibraryType)
        answer = [str(lib_type) for lib_type in sort_list]
        assert gold == answer



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from enum import Enum, unique


@unique
class Lattice(Enum):
    """Represents the options for the lattice ``LAT``."""

    HEXAHEDRA = 1
    """Hexhedra are solids with six faces.

    One such solid is a rectangular prism.
    """
    HEXAGONAL = 2
    """Hexagonal prism are solids with eight faces."""



from hypothesis import given, assume, settings
import hypothesis.strategies as st

import montepy

import pytest

geom_pair = st.tuples(st.integers(min_value=1), st.booleans())


@settings(max_examples=50, deadline=500)
@given(
    st.integers(min_value=1),
    st.lists(geom_pair, min_size=1, max_size=10, unique_by=lambda x: x[0]),
)
def test_build_arbitrary_cell_geometry(first_surf, new_surfaces):
    assume(
        len({first_surf, *[num for num, _ in new_surfaces]}) == len(new_surfaces) + 1
    )
    input = montepy.input_parser.mcnp_input.Input(
        [f"1 0 {first_surf} imp:n=1"], montepy.input_parser.block_type.BlockType.CELL
    )
    cell = montepy.Cell(input)
    surf = montepy.surfaces.surface.Surface()
    surf.number = first_surf
    cell.update_pointers([], [], montepy.surface_collection.Surfaces([surf]))
    for surf_num, operator in new_surfaces:
        surf = montepy.surfaces.surface.Surface()
        surf.number = surf_num
        if operator:
            cell.geometry &= +surf
        else:
            cell.geometry |= -surf
    assert len(cell.geometry) == len(new_surfaces) + 1


def test_cell_geometry_set_warns():
    input = montepy.input_parser.mcnp_input.Input(
        [f"1 0 -1 imp:n=1"], montepy.input_parser.block_type.BlockType.CELL
    )
    cell = montepy.Cell(input)
    with pytest.raises(montepy.errors.IllegalState):
        surf = montepy.surfaces.surface.Surface()
        surf.number = 5
        cell.geometry &= +surf


def test_geom_invalid():
    surf = montepy.AxisPlane()
    with pytest.raises(montepy.errors.IllegalState):
        -surf
    with pytest.raises(montepy.errors.IllegalState):
        +surf
    with pytest.raises(montepy.errors.IllegalState):
        ~montepy.Cell()



class Isotope:
    """A class to represent an MCNP isotope

    .. deprecated:: 0.4.1

        This will class is deprecated, and has been renamed: :class:`~montepy.data_inputs.nuclide.Nuclide`.
        For more details see the :ref:`migrate 0 1`.

    Raises
    ------
    DeprecationWarning
        Whenever called.
    """

    def __init__(self, *args, **kwargs):
        raise DeprecationWarning(
            "montepy.data_inputs.isotope.Isotope is deprecated and is renamed: Nuclide.\n"
            "See <https://www.montepy.org/migrations/migrate0_1.html> for more information "
        )



# Copyright 2024-2025, Battelle Energy Alliance, LLC All Rights Reserved.
from __future__ import annotations
from abc import abstractmethod
import copy
import itertools
from typing import Union
from numbers import Integral

from montepy.mcnp_object import MCNP_Object, InitInput
import montepy
from montepy.utilities import *


def _number_validator(self, number):
    if number < 0:
        raise ValueError("number must be >= 0")
    if self._problem:
        obj_map = montepy.MCNP_Problem._NUMBERED_OBJ_MAP
        try:
            collection_type = obj_map[type(self)]
        except KeyError as e:
            found = False
            for obj_class in obj_map:
                if isinstance(self, obj_class):
                    collection_type = obj_map[obj_class]
                    found = True
                    break
            if not found:
                raise e
        collection = getattr(self._problem, collection_type.__name__.lower())
        collection.check_number(number)
        collection._update_number(self.number, number, self)


class Numbered_MCNP_Object(MCNP_Object):
    """An abstract class to represent an mcnp object that has a number.

    .. versionchanged:: 1.0.0

        Added number parameter

    Parameters
    ----------
    input : Union[Input, str]
        The Input syntax object this will wrap and parse.
    parser : MCNP_Parser
        The parser object to parse the input with.
    number : int
        The number to set for this object.
    """

    def __init__(
        self,
        input: InitInput,
        parser: montepy.input_parser.parser_base.MCNP_Parser,
        number: int = None,
    ):
        self._number = self._generate_default_node(int, -1)
        super().__init__(input, parser)
        self._load_init_num(number)

    def _load_init_num(self, number):
        if number is not None:
            if not isinstance(number, Integral):
                raise TypeError(
                    f"Number must be an int. {number} of type {type(number)} given."
                )
            if number < 0:
                raise ValueError(f"Number must be 0 or greater. {number} given.")
            self.number = number

    _CHILD_OBJ_MAP = {}
    """"""

    @make_prop_val_node("_number", Integral, validator=_number_validator)
    def number(self):
        """The current number of the object that will be written out to a new input.

        Returns
        -------
        int
        """
        pass

    @property
    @abstractmethod
    def old_number(self):
        """The original number of the object provided in the input file

        Returns
        -------
        int
        """
        pass

    def _add_children_objs(self, problem):
        """Adds all children objects from self to the given problem.

        This is called from an append_hook in `NumberedObjectCollection`.
        """
        # skip lambda transforms
        filters = {montepy.Transform: lambda transform: not transform.hidden_transform}
        prob_attr_map = montepy.MCNP_Problem._NUMBERED_OBJ_MAP
        for attr_name, obj_class in self._CHILD_OBJ_MAP.items():
            child_collect = getattr(self, attr_name)
            # allow skipping certain items
            if (
                obj_class in filters
                and child_collect
                and not filters[obj_class](child_collect)
            ):
                continue
            if child_collect:
                prob_collect_name = prob_attr_map[obj_class].__name__.lower()
                prob_collect = getattr(problem, prob_collect_name)
                try:
                    # check if iterable
                    iter(child_collect)
                    assert not isinstance(child_collect, MCNP_Object)
                    # ensure isn't a material or something
                    prob_collect.update(child_collect)
                except (TypeError, AssertionError):
                    prob_collect.append(child_collect)

    def clone(self, starting_number=None, step=None):
        """Create a new independent instance of this object with a new number.

        This relies mostly on ``copy.deepcopy``.

        Notes
        -----
        If starting_number, or step are not specified
        :func:`~montepy.numbered_object_collection.NumberedObjectCollection.starting_number`,
        and :func:`~montepy.numbered_object_collection.NumberedObjectCollection.step` are used as default values,
        if this object is tied to a problem.
        For instance a ``Material`` will use ``problem.materials`` default information.
        Otherwise ``1`` will be used as default values


        .. versionadded:: 0.5.0

        Parameters
        ----------
        starting_number : int
            The starting number to request for a new object number.
        step : int
            the step size to use to find a new valid number.

        Returns
        -------
        type(self)
            a cloned copy of this object.
        """
        if not isinstance(starting_number, (Integral, type(None))):
            raise TypeError(
                f"Starting_number must be an int. {type(starting_number)} given."
            )
        if not isinstance(step, (Integral, type(None))):
            raise TypeError(f"step must be an int. {type(step)} given.")
        if starting_number is not None and starting_number <= 0:
            raise ValueError(f"starting_number must be >= 1. {starting_number} given.")
        if step is not None and step <= 0:
            raise ValueError(f"step must be >= 1. {step} given.")
        ret = copy.deepcopy(self)
        if self._problem:
            ret.link_to_problem(self._problem)
            collection_type = montepy.MCNP_Problem._NUMBERED_OBJ_MAP[type(self)]
            collection = getattr(self._problem, collection_type.__name__.lower())
            if starting_number is None:
                starting_number = collection.starting_number
            if step is None:
                step = collection.step
            ret.number = collection.request_number(starting_number, step)
            collection.append(ret)
            return ret
        if starting_number is None:
            starting_number = 1
        if step is None:
            step = 1
        for number in itertools.count(starting_number, step):
            # only reached if not tied to a problem
            ret.number = number
            if number != self.number:
                return ret



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from .surface_type import SurfaceType
from .surface import Surface, InitInput
from montepy.errors import *
from montepy.utilities import *


def _enforce_positive_radius(self, value):
    if value < 0.0:
        raise ValueError(f"Radius must be positive. {value} given")


class CylinderOnAxis(Surface):
    """Represents surfaces: CX, CY, CZ

    .. versionchanged:: 1.0.0

        Added number parameter

    Parameters
    ----------
    input : Union[Input, str]
        The Input object representing the input
    number : int
        The number to set for this object.
    """

    def __init__(self, input: InitInput = None, number: int = None):
        self._radius = self._generate_default_node(float, None)
        super().__init__(input, number)
        ST = SurfaceType
        if input:
            if self.surface_type not in [ST.CX, ST.CY, ST.CZ]:
                raise ValueError("CylinderOnAxis must be of surface_type: CX, CY, CZ")
            if len(self.surface_constants) != 1:
                raise ValueError("CylinderOnAxis only accepts one surface_constant")
            self._radius = self._surface_constants[0]
        else:
            self._surface_constants = [self._radius]

    @make_prop_val_node(
        "_radius", (float, int), float, validator=_enforce_positive_radius
    )
    def radius(self):
        """The radius of the cylinder

        Returns
        -------
        float
        """
        pass

    def validate(self):
        super().validate()
        if self.radius is None:
            raise IllegalState(f"Surface: {self.number} does not have a radius set.")

    def find_duplicate_surfaces(self, surfaces, tolerance):
        ret = []
        # do not assume transform and periodic surfaces are the same.
        if not self.old_periodic_surface:
            for surface in surfaces:
                if surface != self and surface.surface_type == self.surface_type:
                    if not surface.old_periodic_surface:
                        if abs(self.radius - surface.radius) < tolerance:
                            if self.transform:
                                if surface.transform:
                                    if self.transform.equivalent(
                                        surface.transform, tolerance
                                    ):
                                        ret.append(surface)
                            else:
                                if surface.transform is None:
                                    ret.append(surface)
            return ret
        else:
            return []



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from montepy.surfaces.axis_plane import AxisPlane
from montepy.surfaces.surface import Surface, InitInput
from montepy.surfaces.surface_type import SurfaceType
from montepy.surfaces.cylinder_on_axis import CylinderOnAxis
from montepy.surfaces.cylinder_par_axis import CylinderParAxis
from montepy.surfaces.general_plane import GeneralPlane


def parse_surface(input: InitInput):
    """Builds a Surface object for the type of Surface

    Parameters
    ----------
    input : Union[Input, str]
        The Input object representing the input

    Returns
    -------
    Surface
        A Surface object properly parsed. If supported a sub-class of
        Surface will be given.
    """
    ST = SurfaceType
    buffer_surface = Surface(input)
    type_of_surface = buffer_surface.surface_type
    if type_of_surface in [ST.C_X, ST.C_Y, ST.C_Z]:
        return CylinderParAxis(input)
    elif type_of_surface in [ST.CX, ST.CY, ST.CZ]:
        return CylinderOnAxis(input)
    elif type_of_surface in [ST.PX, ST.PY, ST.PZ]:
        return AxisPlane(input)
    elif type_of_surface == ST.P:
        return GeneralPlane(input)
    else:
        return buffer_surface


surface_builder = parse_surface
"""Alias for :func:`parse_surface`.

:deprecated: 1.0.0
    Renamed to be :func:`parse_surface` to be more pythonic.

Parameters
----------
input : Union[Input, str]
    The Input object representing the input

Returns
-------
Surface
    A Surface object properly parsed. If supported a sub-class of
    Surface will be given.
"""



# Copyright 2024-2025, Battelle Energy Alliance, LLC All Rights Reserved.
from montepy.constants import MAX_ATOMIC_SYMBOL_LENGTH
from montepy._singleton import SingletonGroup
from montepy.data_inputs.element import Element
from montepy.utilities import *
from montepy.input_parser.syntax_node import PaddingNode, ValueNode
from montepy.particle import LibraryType

from functools import total_ordering
import re
from typing import Union
from numbers import Real, Integral

DEFAULT_NUCLIDE_WIDTH = 11
"""How many characters wide a nuclide with spacing should be."""


@total_ordering
class Library(SingletonGroup):
    """A class to represent an MCNP nuclear data library, e.g., ``80c``.

    Examples
    ^^^^^^^^

    .. testcode:: python

        import montepy
        library = montepy.Library("710nc")
        assert library.library == "710nc"
        assert str(library) == "710nc"
        assert library.library_type == montepy.LibraryType.NEUTRON
        assert library.number == 710
        assert library.suffix == "c"

    .. Note::

        This class is immutable, and hashable, meaning it is suitable as a dictionary key.

    .. versionadded:: 1.0.0

    Parameters
    ----------
    library : str
        The name of the library.

    Raises
    ------
    TypeErrror
        if a string is not provided.
    ValueError
        if a valid library is not provided.
    """

    __slots__ = "_library", "_lib_type", "_num", "_suffix"

    _SUFFIX_MAP = {
        "c": LibraryType.NEUTRON,
        "d": LibraryType.NEUTRON,
        "m": LibraryType.NEUTRON,  # coupled neutron photon, invokes `g`
        "g": LibraryType.PHOTO_ATOMIC,
        "p": LibraryType.PHOTO_ATOMIC,
        "u": LibraryType.PHOTO_NUCLEAR,
        "y": LibraryType.NEUTRON,
        "e": LibraryType.ELECTRON,
        "h": LibraryType.PROTON,
        "o": LibraryType.DEUTERON,
        "r": LibraryType.TRITON,
        "s": LibraryType.HELION,
        "a": LibraryType.ALPHA_PARTICLE,
    }
    _LIBRARY_RE = re.compile(r"(\d{2,3})[a-z]?([a-z])", re.I)

    def __init__(self, library: str):
        self._lib_type = None
        self._suffix = ""
        self._num = None
        if not isinstance(library, str):
            raise TypeError(f"library must be a str. {library} given.")
        if library:
            match = self._LIBRARY_RE.fullmatch(library)
            if not match:
                raise ValueError(f"Not a valid library. {library} given.")
            self._num = int(match.group(1))
            extension = match.group(2).lower()
            self._suffix = extension
            try:
                lib_type = self._SUFFIX_MAP[extension]
            except KeyError:
                raise ValueError(
                    f"Not a valid library extension suffix. {library} with extension: {extension} given."
                )
            self._lib_type = lib_type
        self._library = library

    @property
    def library(self) -> str:
        """The full name of the library.

        Returns
        -------
        str
        """
        return self._library

    @property
    def library_type(self) -> LibraryType:
        """The :class:`~montepy.particle.LibraryType` of this library.

        This corresponds to the type of library this would specified
        in a material definition e.g., ``NLIB``, ``PLIB``, etc.

        See Also
        --------

        * :manual63:`5.6.1`

        Returns
        -------
        LibraryType
            the type of library this library is.
        """
        return self._lib_type

    @property
    def number(self) -> int:
        """The base number in the library.

        For example: this would be ``80`` for the library: ``Library('80c')``.

        Returns
        -------
        int
            the base number of the library.
        """
        return self._num

    @property
    def suffix(self) -> str:
        """The suffix of the library, or the final character of its definition.

        For example this would be ``"c"`` for the library: ``Library('80c')``.

        Returns
        -------
        str
            the suffix of the library.
        """
        return self._suffix

    def __hash__(self):
        return hash(self._library.upper())

    def __eq__(self, other):
        if not isinstance(other, (type(self), str)):
            raise TypeError(f"Can only compare Library instances.")
        if not isinstance(other, type(self)):
            return self.library.upper() == other.upper()
        # due to SingletonGroup
        return self.library.upper() == other.library.upper()

    def __bool__(self):
        return bool(self.library)

    def __str__(self):
        return self.library

    def __repr__(self):
        return f"Library('{self.library}')"

    def __lt__(self, other):
        if not isinstance(other, (str, type(self))):
            raise TypeError(f"Can only compare Library instances.")
        if isinstance(other, str):
            other = Library(other)
        if self.suffix == other.suffix:
            return self.number < other.number
        return self.suffix < other.suffix

    def __reduce__(self):
        return (self.__class__, (self._library,))


_ZAID_A_ADDER = 1000
"""How much to multiply Z by to form a ZAID."""


class Nucleus(SingletonGroup):
    """A class to represent a nuclide irrespective of the nuclear data being used.

    This is meant to be an immutable representation of the nuclide, no matter what nuclear data
    library is used. ``U-235`` is always ``U-235``.
    Generally users don't need to interact with this much as it is almost always wrapped
    by: :class:`montepy.data_inputs.nuclide.Nuclide`.

    .. Note::

        This class is immutable, and hashable, meaning it is suitable as a dictionary key.

    .. versionadded:: 1.0.0

    Parameters
    ----------
    element : Element
        the element this Nucleus is based on.
    A : int
        The A-number (atomic mass) of the nuclide. If this is elemental
        this should be 0.
    meta_state : int
        The metastable state if this nuclide is isomer.

    Raises
    ------
    TypeError
        if an parameter is the wrong type.
    ValueError
        if non-sensical values are given.
    """

    __slots__ = "_element", "_A", "_meta_state"

    def __init__(
        self,
        element: Element,
        A: int = 0,
        meta_state: int = 0,
    ):
        if not isinstance(element, Element):
            raise TypeError(
                f"Only type Element is allowed for element argument. {element} given."
            )
        self._element = element

        if not isinstance(A, Integral):
            raise TypeError(f"A number must be an int. {A} given.")
        if A < 0:
            raise ValueError(f"A cannot be negative. {A} given.")
        self._A = A
        if not isinstance(meta_state, (Integral, type(None))):
            raise TypeError(f"Meta state must be an int. {meta_state} given.")
        if A == 0 and meta_state != 0:
            raise ValueError(
                f"A metastable elemental state is Non-sensical. A: {A}, meta_state: {meta_state} given."
            )
        if meta_state not in range(0, 5):
            raise ValueError(
                f"Meta state can only be in the range: [0,4]. {meta_state} given."
            )
        self._meta_state = meta_state

    @property
    def ZAID(self) -> int:
        """The ZZZAAA identifier following MCNP convention.

        If this is metastable the MCNP convention for ZAIDs for metastable isomers will be used.

        Returns
        -------
        int
        """
        meta_adder = 300 + 100 * self.meta_state if self.is_metastable else 0
        temp = self.Z * _ZAID_A_ADDER + self.A + meta_adder
        if temp in Nuclide._STUPID_ZAID_SWAP:
            return Nuclide._STUPID_ZAID_SWAP[temp]
        return temp

    @property
    def Z(self) -> int:
        """The Z number for this isotope.

        Returns
        -------
        int
            the atomic number.
        """
        return self._element.Z

    @make_prop_pointer("_A")
    def A(self) -> int:
        """The A number for this isotope.

        Returns
        -------
        int
            the isotope's mass.
        """
        pass

    @make_prop_pointer("_element")
    def element(self) -> Element:
        """The base element for this isotope.

        Returns
        -------
        Element
            The element for this isotope.
        """
        pass

    @property
    def is_metastable(self) -> bool:
        """Whether or not this is a metastable isomer.

        Returns
        -------
        bool
            boolean of if this is metastable.
        """
        return bool(self._meta_state)

    @make_prop_pointer("_meta_state")
    def meta_state(self) -> int:
        """If this is a metastable isomer, which state is it?

        Can return values in the range [0,4]. The exact state
        number is decided by who made the ACE file for this, and not quantum mechanics.
        Convention states that the isomers should be numbered from lowest to highest energy.
        The ground state will be 0.

        Returns
        -------
        int
            the metastable isomeric state of this "isotope" in the range
            [0,4].
        """
        pass

    def __hash__(self):
        return hash((self.element, self.A, self.meta_state))

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            raise TypeError(
                f"Nucleus can only be compared to a Nucleus. {other} of type {type(other)} given."
            )
        # due to SingletonGroup
        return (
            self.element == other.element
            and self.A == other.A
            and self.meta_state == other.meta_state
        )

    def __reduce__(self):
        return (type(self), (self.element, self.A, self._meta_state))

    def __lt__(self, other):
        if not isinstance(other, type(self)):
            raise TypeError
        return (self.Z, self.A, self.meta_state) < (other.Z, other.A, other.meta_state)

    def __str__(self):
        meta_suffix = f"m{self.meta_state}" if self.is_metastable else ""
        return f"{self.element.symbol:>2}-{self.A:<3}{meta_suffix:<2}"

    def __repr__(self):
        return f"Nucleus({self.element}, {self.A}, {self.meta_state})"


class Nuclide:
    r"""A class to represent an MCNP nuclide with nuclear data library information.

    Nuclide accepts ``name`` as a way of specifying a nuclide.
    This is meant to be more ergonomic than ZAIDs while not going insane with possible formats.
    This accepts ZAID and Atomic_symbol-A format.
    All cases support metastables as m# and a library specification.
    Examples include:

    * ``1001.80c``
    * ``92235m1.80c``
    * ``92635.80c``
    * ``U.80c``
    * ``U-235.80c``
    * ``U-235m1.80c``

    To be specific this must match the regular expression:

    .. code-block::

        import re
        parser = re.compile(\"\"\"
            (\d{4,6}) # ZAID
                |
            ([a-z]{1,2} # or atomic symbol
            -?\d*) # optional A-number
            (m\d+)? # optional metastable
            (\.\d{{2,}}[a-z]+)? # optional library
            \"\"\",
            re.IGNORE_CASE | re.VERBOSE
        )

    .. Note::

        As discussed in :manual63:`5.6.1`:

            To represent a metastable isotope, adjust the AAA value using the
            following convention: AAA’=(AAA+300)+(m × 100), where m is the
            metastable level and m=1, 2, 3, or 4.

        MontePy attempts to apply these rules to determine the isomeric state of the nuclide.
        This requires MontePy to determine if a ZAID is a realistic base isomeric state.

        This is done simply by manually specifying 6 rectangles of realistic ZAIDs.
        MontePy checks if a ZAID is inside of these rectangles.
        These rectangles are defined by their upper right corner as an isotope.
        The lower left corner is defined by the Z-number of the previous isotope and A=0.

        These isotopes are:

        * Cl-52
        * Br-101
        * Xe-150
        * Os-203
        * Cm-251
        * Og-296

    .. Warning::

        Due to legacy reasons the nuclear data for Am-242 and Am-242m1 have been swapped for the nuclear data
        provided by LANL.
        This is documented in `section 1.2.2 of the MCNP 6.3.1 manual <https://www.osti.gov/servlets/purl/2372634>`_ :

            As a historical quirk, 242m1Am and 242Am are swapped in the ZAID and SZAID formats, so that the
            former is 95242 and the latter is 95642 for ZAID and 1095242 for SZAID. It is important to verify if a
            data library follows this convention. To date, all LANL-published libraries do. The name format does
            not swap these isomers. As such, Am-242m1 can load a table labeled 95242.

        Due to this MontePy follows the MCNP convention, and swaps these ZAIDs.
        If you have custom generated ACE data for Am-242,
        that does not follow this convention you have a few options:

        #. Do nothing. If you do not need to modify a material in an MCNP input file the ZAID will be written out the same as it was in the original file.

        #. Specify the Nucleus by ZAID. This will have the same effect as before. Note that MontePy will display the wrong metastable state, but will preserve the ZAID.

        #. Open an issue. If this approach doesn't work for you please open an issue so we can develop a better solution.

    See Also
    --------

    * :manual62:`107`
    * :manual63:`5.6.1`


    .. versionadded:: 1.0.0

        This was added as replacement for ``montepy.data_inputs.Isotope``.

    Parameters
    ----------
    name : str
        A fancy name way of specifying a nuclide.
    ZAID : str
        The ZAID in MCNP format, the library can be included.
    element : Element
        the element this Nucleus is based on.
    Z : int
        The Z-number (atomic number) of the nuclide.
    A : int
        The A-number (atomic mass) of the nuclide. If this is elemental
        this should be 0.
    meta_state : int
        The metastable state if this nuclide is isomer.
    library : str
        the library to use for this nuclide.
    node : ValueNode
        The ValueNode to build this off of. Should only be used by
        MontePy.

    Raises
    ------
    TypeError
        if a parameter is the wrong type.
    ValueError
        if non-sensical values are given.
    """

    _NAME_PARSER = re.compile(
        rf"""(
                (?P<ZAID>\d{{4,6}})|
                ((?P<element>[a-z]{{1,{MAX_ATOMIC_SYMBOL_LENGTH}}})-?(?P<A>\d*))
            )
            (m(?P<meta>\d+))?
            (\.(?P<library>\d{{2,}}[a-z]+))?""",
        re.I | re.VERBOSE,
    )
    """Parser for fancy names."""

    #                   Cl-52      Br-101     Xe-150      Os-203    Cm-251     Og-296
    _BOUNDING_CURVE = [(17, 52), (35, 101), (54, 150), (76, 203), (96, 251), (118, 296)]
    """Points on bounding curve for determining if "valid" isotope"""
    _STUPID_MAP = {
        "95642": {"_meta_state": 0},
        "95242": {"_meta_state": 1},
    }
    _STUPID_ZAID_SWAP = {95242: 95642, 95642: 95242}

    def __init__(
        self,
        name: Union[str, int, Element, Nucleus] = "",
        element: Element = None,
        Z: int = None,
        A: int = 0,
        meta_state: int = 0,
        library: str = "",
        node: ValueNode = None,
    ):
        self._library = Library("")
        ZAID = ""

        if not isinstance(name, (str, Integral, Element, Nucleus, Nuclide, type(None))):
            raise TypeError(
                f"Name must be str, int, Element, or Nucleus. {name} of type {type(name)} given."
            )
        if name:
            element, A, meta_state, new_library = self._parse_fancy_name(name)
            # give library precedence always
            if library == "":
                library = new_library
        if node is not None and isinstance(node, ValueNode):
            if node.type == float:
                node = ValueNode(node.token, str, node.padding)
            self._tree = node
            ZAID = node.value
        parts = ZAID.split(".")
        if ZAID:
            za_info = self._parse_zaid(int(parts[0]))
            element = za_info["_element"]
            A = za_info["_A"]
            meta_state = za_info["_meta_state"]
        if Z:
            element = Element(Z)
        if element is None:
            raise ValueError(
                "no elemental information was provided via name, element, or z. "
                f"Given: name: {name}, element: {element}, Z: {Z}"
            )
        self._nucleus = Nucleus(element, A, meta_state)
        if len(parts) > 1 and library == "":
            library = parts[1]
        if not isinstance(library, str):
            raise TypeError(f"Library can only be str. {library} given.")
        self._library = Library(library)
        if not node:
            padding_num = DEFAULT_NUCLIDE_WIDTH - len(self.mcnp_str())
            if padding_num < 1:
                padding_num = 1
            self._tree = ValueNode(self.mcnp_str(), str, PaddingNode(" " * padding_num))

    @classmethod
    def _handle_stupid_legacy_stupidity(cls, ZAID):
        """This handles legacy issues where ZAID are swapped.

        For now this is only for Am-242 and Am-242m1.

        See Also
        --------

        * :manual631:`1.2.2`
        """
        ZAID = str(ZAID)
        ret = {}
        if ZAID in cls._STUPID_MAP:
            stupid_overwrite = cls._STUPID_MAP[ZAID]
            for key, value in stupid_overwrite.items():
                ret[key] = value
        return ret

    @classmethod
    def _parse_zaid(cls, ZAID) -> dict[str, object]:
        """Parses the ZAID fully including metastable isomers.

        See Table 3-32 of LA-UR-17-29881

        Parameters
        ----------
        ZAID : int
            the ZAID without the library

        Returns
        -------
        dict[str, Object]
            a dictionary with the parsed information, in a way that can
            be loaded into nucleus. Keys are: _element, _A, _meta_state
        """

        def is_probably_an_isotope(Z, A):
            for lim_Z, lim_A in cls._BOUNDING_CURVE:
                if Z <= lim_Z:
                    if A <= lim_A:
                        return True
                    else:
                        return False
                else:
                    continue
            # if you are above Og it's probably legit.
            # to reach this state requires new elements to be discovered.
            return True  # pragma: no cover

        ret = {}
        Z = int(ZAID / _ZAID_A_ADDER)
        ret["_element"] = Element(Z)
        ret["_A"] = 0
        ret["_meta_state"] = 0
        A = int(ZAID % _ZAID_A_ADDER)
        ret["_A"] = A
        if not is_probably_an_isotope(Z, A):
            true_A = A - 300
            # only m1,2,3,4 allowed
            found = False
            for i in range(1, 5):
                true_A -= 100
                # assumes that can only vary 40% from A = 2Z
                if is_probably_an_isotope(Z, true_A):
                    found = True
                    break
            if found:
                ret["_meta_state"] = i
                ret["_A"] = true_A
            else:
                raise ValueError(
                    f"ZAID: {ZAID} cannot be parsed as a valid metastable isomer. "
                    "Only isomeric state 0 - 4 are allowed"
                )

        ret.update(cls._handle_stupid_legacy_stupidity(ZAID))
        return ret

    @property
    def ZAID(self) -> int:
        """The ZZZAAA identifier following MCNP convention

        Returns
        -------
        int
        """
        # if this is made mutable this cannot be user provided, but must be calculated.
        return self._nucleus.ZAID

    @property
    def Z(self) -> int:
        """The Z number for this isotope.

        Returns
        -------
        int
            the atomic number.
        """
        return self._nucleus.Z

    @property
    def A(self) -> int:
        """The A number for this isotope.

        Returns
        -------
        int
            the isotope's mass.
        """
        return self._nucleus.A

    @property
    def element(self) -> Element:
        """The base element for this isotope.

        Returns
        -------
        Element
            The element for this isotope.
        """
        return self._nucleus.element

    @make_prop_pointer("_nucleus")
    def nucleus(self) -> Nucleus:
        """The base nuclide of this nuclide without the nuclear data library.

        Returns
        -------
        Nucleus
        """
        pass

    @property
    def is_metastable(self) -> bool:
        """Whether or not this is a metastable isomer.

        Returns
        -------
        bool
            boolean of if this is metastable.
        """
        return self._nucleus.is_metastable

    @property
    def meta_state(self) -> int:
        """If this is a metastable isomer, which state is it?

        Can return values in the range [0,4]. 0 corresponds to the ground state.
        The exact state number is decided by who made the ACE file for this, and not quantum mechanics.
        Convention states that the isomers should be numbered from lowest to highest energy.

        Returns
        -------
        int
            the metastable isomeric state of this "isotope" in the range
            [0,4]l
        """
        return self._nucleus.meta_state

    @make_prop_pointer("_library", (str, Library), Library)
    def library(self) -> Library:
        """The MCNP library identifier e.g. 80c

        Returns
        -------
        Library
        """
        pass

    def mcnp_str(self) -> str:
        """Returns an MCNP formatted representation.

        E.g., 1001.80c

        Returns
        -------
        str
            a string that can be used in MCNP
        """
        return f"{self.ZAID}.{self.library}" if str(self.library) else str(self.ZAID)

    def nuclide_str(self) -> str:
        """Creates a human readable version of this nuclide excluding the data library.

        This is of the form Atomic symbol - A [metastable state]. e.g., ``U-235m1``.

        Returns
        -------
        str
        """
        meta_suffix = f"m{self.meta_state}" if self.is_metastable else ""
        suffix = f".{self._library}" if str(self._library) else ""
        return f"{self.element.symbol}-{self.A}{meta_suffix}{suffix}"

    def get_base_zaid(self) -> int:
        """Get the ZAID identifier of the base isotope this is an isomer of.

        This is mostly helpful for working with metastable isomers.

        Returns
        -------
        int
            the mcnp ZAID of the ground state of this isotope.
        """
        return self.Z * _ZAID_A_ADDER + self.A

    @classmethod
    def _parse_fancy_name(cls, identifier):
        """Parses a fancy name that is a ZAID, a Symbol-A, or nucleus, nuclide, or element.

        Parameters
        ----------
        identifier
        idenitifer : Union[str, int, element, Nucleus, Nuclide]

        Returns
        -------
        tuple
            a tuple of element, a, isomer, library
        """
        if isinstance(identifier, (Nucleus, Nuclide)):
            if isinstance(identifier, Nuclide):
                lib = identifier.library
            else:
                lib = ""
            return (identifier.element, identifier.A, identifier.meta_state, str(lib))
        if isinstance(identifier, Element):
            element = identifier
            return (element, 0, 0, "")
        A = 0
        isomer = 0
        library = ""
        if isinstance(identifier, Real):
            if identifier > _ZAID_A_ADDER:
                parts = Nuclide._parse_zaid(int(identifier))
                element, A, isomer = (
                    parts["_element"],
                    parts["_A"],
                    parts["_meta_state"],
                )
            else:
                element = Element(int(identifier))
        elif isinstance(identifier, str):
            if match := cls._NAME_PARSER.fullmatch(identifier):
                match = match.groupdict()
                if match["ZAID"]:
                    parts = cls._parse_zaid(int(match["ZAID"]))
                    element, A, isomer = (
                        parts["_element"],
                        parts["_A"],
                        parts["_meta_state"],
                    )
                else:
                    element_name = match["element"]
                    element = Element.get_by_symbol(element_name.capitalize())
                    if match["A"]:
                        A = int(match["A"])
                if match["meta"]:
                    extra_isomer = int(match["meta"])
                    isomer += extra_isomer
                if match["library"]:
                    library = match["library"]
            else:
                raise ValueError(f"Not a valid nuclide identifier. {identifier} given")

        return (element, A, isomer, library)

    def __repr__(self):
        return f"{self.__class__.__name__}({repr(self.nuclide_str())})"

    def __str__(self):
        meta_suffix = f"m{self.meta_state}" if self.is_metastable else ""
        suffix = f" ({self._library})" if str(self._library) else "()"
        return f"{self.element.symbol:>2}-{self.A:<3}{meta_suffix:<2}{suffix:>5}"

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            raise TypeError(
                f"Cannot compare Nuclide to other values. {other} of type {type(other)}."
            )
        return self.nucleus == other.nucleus and self.library == other.library

    def __lt__(self, other):
        if not isinstance(other, type(self)):
            raise TypeError(
                f"Cannot compare Nuclide to other values. {other} of type {type(other)}."
            )
        return (self.nucleus, self.library) < (other.nucleus, other.library)

    def __format__(self, format_str):
        return str(self).__format__(format_str)



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from . import axis_plane
from . import cylinder_par_axis
from . import cylinder_on_axis
from . import half_space
from . import surface
from . import surface_builder

# promote objects
from .axis_plane import AxisPlane
from .cylinder_par_axis import CylinderParAxis
from .cylinder_on_axis import CylinderOnAxis
from .half_space import HalfSpace, UnitHalfSpace
from .surface import Surface
from .surface_type import SurfaceType

# promote functions
from .surface_builder import parse_surface



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
import itertools as it
from montepy.constants import ASCII_CEILING
from montepy.utilities import *
import os


class MCNP_InputFile:
    """A class to represent a distinct input file.

    .. Note::
        this is a bare bones implementation to be fleshed out in the future.

    .. versionchanged:: 0.3.0
        Added the overwrite attribute.

    Parameters
    ----------
    path : str
        the path to the input file
    parent_file : str
        the parent file for this file if any. This occurs when a "read"
        input is used.
    overwrite : bool
        Whether to overwrite the file 'path' if it exists
    """

    def __init__(self, path, parent_file=None, overwrite=False):
        self._path = path
        self._parent_file = parent_file
        self._lineno = 1
        self._replace_with_space = False
        self._overwrite = overwrite
        self._mode = None
        self._fh = None
        self._is_stream = False

    @classmethod
    def from_open_stream(cls, fh):
        """Create an MCNP Input File from an open, writable stream

        Parameters
        ----------
        fh : io.TextIOBase
            An open and writable object, such as a file handle.
        """
        name = getattr(fh, "name", fh.__class__.__name__)
        inpfile = cls(path=name)
        inpfile._fh = fh
        inpfile._is_stream = True
        return inpfile

    @make_prop_pointer("_path")
    def path(self):
        """The path for the file.

        Returns
        -------
        str
        """
        pass

    @property
    def name(self):
        return self._path

    @property
    def is_stream(self):
        return self._is_stream

    @make_prop_pointer("_parent_file")
    def parent_file(self):
        """The parent file for this file.

        This is only used when this file is pointed to by a "read" input.

        Returns
        -------
        str
        """
        pass

    @make_prop_pointer("_lineno")
    def lineno(self):
        """The current line number being read in the file.

        This is 1-indexed.

        Returns
        -------
        int
        """
        pass

    def open(self, mode, encoding="ascii", replace=True):
        """Opens the underlying file, and returns self.

        This should only ever be completed from within a ``with`` statement.
        For this reason, a ``close`` functional is intentionally
        not provided.

        .. Note::
            For different encoding schemes see the available list
            `here <https://docs.python.org/3.9/library/codecs.html#standard-encodings>`_.

            CP1252 is commonly referred to as "extended-ASCII".
            You may have success with this encoding for working with special characters.

        .. versionchanged:: 0.2.11
            Added guardrails to raise FileExistsError and IsADirectoryError.

        Parameters
        ----------
        mode : str
            the mode to open the file in
        encoding : str
            The encoding scheme to use. If replace is true, this is
            ignored, and changed to ASCII
        replace : bool
            replace all non-ASCII characters with a space (0x20)

        Returns
        -------
        unknown
            self

        Raises
        ------
        FileExistsError
            if a file already exists with the same path while writing.
        IsADirectoryError
            if the path given is actually a directory while writing.
        """
        if "r" in mode:
            if replace:
                self._replace_with_space = True
                mode = "rb"
                encoding = None
        self._mode = mode
        if "w" in mode:
            if os.path.isfile(self.path) and self._overwrite is not True:
                raise FileExistsError(
                    f"{self.path} already exists, and overwrite is not set."
                )
            if os.path.isdir(self.path):
                raise IsADirectoryError(
                    f"{self.path} is a directory, and cannot be overwritten."
                )
        self._fh = open(self.path, mode, encoding=encoding)
        return self

    def __enter__(self):
        self._fh.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        status = self._fh.__exit__(exc_type, exc_val, exc_tb)
        self._fh = None
        return status

    def __iter__(self):
        for lineno, line in enumerate(self._fh):
            self._lineno = lineno + 1
            if self._mode == "rb" and self._replace_with_space:
                line = self._clean_line(line)
            yield line

    @staticmethod
    def _clean_line(line):
        new_line = bytes([code if code < ASCII_CEILING else ord(" ") for code in line])
        line = new_line.decode("ascii")
        line = line.replace("\r\n", "\n").replace("\r", "\n")
        return line

    def read(self, size=-1):
        """"""
        if self._fh:
            ret = self._fh.read(size)
            if self._mode == "rb" and self._replace_with_space:
                ret = self._clean_line(ret)
            self._lineno += ret.count("\n")
            return ret

    def readline(self, size=-1):
        """"""
        if self._fh:
            ret = self._fh.readline(size)
            if self._mode == "rb" and self._replace_with_space:
                ret = self._clean_line(ret)
            self._lineno += ret.count("\n")
            return ret

    def write(self, to_write):
        if self._fh:
            self._lineno += to_write.count("\n")
            return self._fh.write(to_write)

    def __str__(self):
        return str(self.name)



# Copyright 2024-2025, Battelle Energy Alliance, LLC All Rights Reserved.
import montepy
from montepy.cells import Cells
from montepy.input_parser.mcnp_input import Input
from montepy.input_parser.block_type import BlockType
from montepy.input_parser import syntax_node
from montepy.numbered_mcnp_object import Numbered_MCNP_Object

from numbers import Integral


class Universe(Numbered_MCNP_Object):
    """Class to represent an MCNP universe, but not handle the input
    directly.

    Parameters
    ----------
    number : int
        The number for the universe, must be ≥ 0
    """

    def __init__(self, number: int):
        self._number = self._generate_default_node(int, -1)
        if not isinstance(number, Integral):
            raise TypeError("number must be int")
        if number < 0:
            raise ValueError(f"Universe number must be ≥ 0. {number} given.")
        self._number = self._generate_default_node(int, number)

        class Parser:
            def parse(self, token_gen, input):
                return syntax_node.SyntaxNode("fake universe", {})

        super().__init__(Input(["U"], BlockType.DATA), Parser(), number)

    @property
    def cells(self):
        """A generator of the cell objects in this universe.

        Returns
        -------
        Generator
            a generator returning every cell in this universe.
        """
        if self._problem:
            for cell in self._problem.cells:
                if cell.universe == self:
                    yield cell

    def claim(self, cells):
        """Take the given cells and move them into this universe, and out of their original universe.

        Can be given a single Cell, a list of cells, or a Cells object.

        Parameters
        ----------
        cells : Cell, list, or Cells
            the cell(s) to be claimed

        Raises
        ------
        TypeError
            if bad parameter is given.
        """
        if not isinstance(cells, (montepy.Cell, list, Cells)):
            raise TypeError(f"Cells being claimed must be a Cell, list, or Cells")
        if isinstance(cells, list):
            cells = Cells(cells)
        if isinstance(cells, montepy.Cell):
            cells = Cells([cells])

        for cell in cells:
            cell.universe = self

    @property
    def old_number(self):
        """Original universe number from the input file."""
        return self._number

    def _update_values(self):
        pass

    def __str__(self):
        return f"Universe({self.number})"

    def __repr__(self):
        return (
            f"Universe: Number: {self.number} "
            f"Problem: {'set' if self._problem else 'not set'}, "
            f"Cells: {[cell.number for cell in self.cells] if self._problem else ''}"
        )



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
import hypothesis
from hypothesis import given, settings, strategies as st
import copy
import itertools as it

import montepy
import montepy.cells
from montepy.errors import NumberConflictError
import pytest
import os


class TestNumberedObjectCollection:

    @pytest.fixture(scope="class")
    def read_simple_problem(_):
        return montepy.read_input(os.path.join("tests", "inputs", "test.imcnp"))

    @pytest.fixture
    def cp_simple_problem(_, read_simple_problem):
        return copy.deepcopy(read_simple_problem)

    def test_bad_init(self):
        with pytest.raises(TypeError):
            montepy.Cells(5)
        with pytest.raises(TypeError):
            montepy.Cells([5])

    def test_numbers(self, cp_simple_problem):
        cell_numbers = [1, 2, 3, 99, 5]
        surf_numbers = [1000, 1005, 1010, 1015, 1020, 1025]
        mat_numbers = [1, 2, 3]
        problem = cp_simple_problem
        assert list(problem.cells.numbers) == cell_numbers
        assert list(problem.surfaces.numbers) == surf_numbers
        assert list(problem.materials.numbers) == mat_numbers

    def test_number_conflict_init(self, cp_simple_problem):
        cells = list(cp_simple_problem.cells)
        cells.append(cells[1])
        with pytest.raises(NumberConflictError):
            montepy.cells.Cells(cells)

    def test_check_number(self, cp_simple_problem):
        with pytest.raises(NumberConflictError):
            cp_simple_problem.cells.check_number(1)
        with pytest.raises(TypeError):
            cp_simple_problem.cells.check_number("5")
        # testing a number that shouldn't conflict to ensure error isn't raised
        cp_simple_problem.cells.check_number(20)

    def test_objects(self, cp_simple_problem):
        generated = list(cp_simple_problem.cells)
        objects = cp_simple_problem.cells.objects
        assert generated == objects

    def test_pop(self, cp_simple_problem):
        cells = copy.deepcopy(cp_simple_problem.cells)
        size = len(cells)
        target = list(cells)[-1]
        popped = cells.pop()
        assert target == popped
        assert size - 1 == len(cells)
        with pytest.raises(TypeError):
            cells.pop("hi")

    def test_extend(self, cp_simple_problem):
        surfaces = copy.deepcopy(cp_simple_problem.surfaces)
        extender = list(surfaces)[0:2]
        size = len(surfaces)
        with pytest.raises(NumberConflictError):
            surfaces.extend(extender)
        assert len(surfaces) == size
        extender = copy.deepcopy(extender)
        extender[0].number = 50
        extender[1].number = 60
        surfaces.extend(extender)
        assert len(surfaces) == size + 2
        # force a num_cache miss
        extender = copy.deepcopy(extender)
        for surf in extender:
            surf._problem = None
        surfaces[1000].number = 1
        extender[0].number = 1000
        extender[1].number = 70
        surfaces.extend(extender)
        assert len(surfaces) == size + 4
        with pytest.raises(TypeError):
            surfaces.extend(5)
        with pytest.raises(TypeError):
            surfaces.extend([5])

    def test_iter(self, cp_simple_problem):
        size = len(cp_simple_problem.cells)
        counter = 0
        for cell in cp_simple_problem.cells:
            counter += 1
        assert size == counter

    def test_append(self, cp_simple_problem):
        cells = copy.deepcopy(cp_simple_problem.cells)
        cell = copy.deepcopy(cells[1])
        size = len(cells)
        with pytest.raises(NumberConflictError):
            cells.append(cell)
        with pytest.raises(TypeError):
            cells.append(5)
        cell.number = 20
        cells.append(cell)
        assert len(cells) == size + 1

    def test_add(_):
        cells = montepy.Cells()
        cell = montepy.Cell()
        cell.number = 2
        cells.add(cell)
        assert cell in cells
        # test silent no-op
        cells.add(cell)
        cell = copy.deepcopy(cell)
        with pytest.raises(NumberConflictError):
            cells.add(cell)
        with pytest.raises(TypeError):
            cells.add(5)

    def test_update(_):
        cells = montepy.Cells()
        cell_list = []
        for i in range(1, 6):
            cell_list.append(montepy.Cell())
            cell_list[-1].number = i
        cells.update(cell_list)
        for cell in cell_list:
            assert cell in cells
        with pytest.raises(TypeError):
            cells.update(5)
        with pytest.raises(TypeError):
            cells.update({5})
        cell = montepy.Cell()
        cell.number = 1
        cells.update([cell])
        assert cells[1] is cell_list[0]
        assert cells[1] is not cell

    def test_append_renumber(self, cp_simple_problem):
        cells = copy.deepcopy(cp_simple_problem.cells)
        size = len(cells)
        cell = copy.deepcopy(cells[1])
        cell.number = 20
        cells.append_renumber(cell)
        assert len(cells) == size + 1
        with pytest.raises(TypeError):
            cells.append_renumber(5)
        with pytest.raises(TypeError):
            cells.append_renumber(cell, "hi")
        cell = copy.deepcopy(cell)
        cell._problem = None
        cell.number = 1
        cells.append_renumber(cell)
        assert cell.number == 4
        assert len(cells) == size + 2

    def test_append_renumber_problems(self, cp_simple_problem):
        print(hex(id(cp_simple_problem.materials._problem)))
        prob1 = copy.deepcopy(cp_simple_problem)
        prob2 = copy.deepcopy(cp_simple_problem)
        print(hex(id(cp_simple_problem.materials._problem)))
        # Delete Material 2, making its number available.
        prob2.materials.remove(prob2.materials[2])
        len_mats = len(prob2.materials)
        mat1 = prob1.materials[1]
        new_num = prob2.materials.append_renumber(mat1)
        assert new_num == 2, "Material not renumbered correctly."
        assert len(prob2.materials) == len_mats + 1, "Material not appended"
        assert prob2.materials[2] is mat1, "Material 2 is not the new material"

    def test_request_number(self, cp_simple_problem):
        cells = cp_simple_problem.cells
        assert cells.request_number(6) == 6
        assert cells.request_number(1) == 4
        assert cells.request_number(99, 6) == 105
        with pytest.raises(TypeError):
            cells.request_number("5")
        with pytest.raises(TypeError):
            cells.request_number(1, "5")

    def test_next_number(self, cp_simple_problem):
        cells = cp_simple_problem.cells
        assert cells.next_number() == 100
        assert cells.next_number(6) == 105
        with pytest.raises(TypeError):
            cells.next_number("5")
        with pytest.raises(ValueError):
            cells.next_number(-1)

    def test_getitem(self, cp_simple_problem):
        cells = cp_simple_problem.cells
        list_version = list(cells)
        assert cells[1] == list_version[0]
        # force stale cache misses
        cells[1].number = 20
        with pytest.raises(KeyError):
            cells[1]
        # force cache miss
        assert cells[20] == list_version[0]
        with pytest.raises(TypeError):
            cells["5"]

    def test_delete(self, cp_simple_problem):
        cells = copy.deepcopy(cp_simple_problem.cells)
        size = len(cells)
        del cells[1]
        assert size - 1 == len(cells)
        with pytest.raises(TypeError):
            del cells["5"]

    def test_setitem(self, cp_simple_problem):
        cells = copy.deepcopy(cp_simple_problem.cells)
        cell = cells[1]
        size = len(cells)
        cell = copy.deepcopy(cell)
        with pytest.raises(NumberConflictError):
            cells[1] = cell
        with pytest.raises(TypeError):
            cells[1] = 5
        with pytest.raises(TypeError):
            cells["1"] = cell
        cell = copy.deepcopy(cell)
        cell.number = 20
        cells[50] = cell
        assert len(cells) == size + 1

    def test_iadd(self, cp_simple_problem):
        cells = copy.deepcopy(cp_simple_problem.cells)
        list_cells = list(cells)
        size = len(cells)
        with pytest.raises(NumberConflictError):
            cells += list_cells
        with pytest.raises(NumberConflictError):
            cells += montepy.cells.Cells(list_cells)

        with pytest.raises(TypeError):
            cells += 5
        with pytest.raises(TypeError):
            cells += [5]

        list_cells = [copy.deepcopy(cells[1])]
        list_cells[0].number = 20
        cells += list_cells
        assert len(cells) == size + 1

        this_problem = copy.deepcopy(cp_simple_problem)
        # just ignore materials being added
        this_problem.materials.clear()
        for cell in this_problem.cells:
            cell.number += 1000
        this_problem.cells += cp_simple_problem.cells
        assert len(this_problem.cells) == size * 2

    def test_slice(self, cp_simple_problem):
        test_numbers = [c.number for c in cp_simple_problem.cells[1:5]]
        assert [1, 2, 3, 5] == test_numbers
        test_numbers = [c.number for c in cp_simple_problem.cells[2:]]
        assert [2, 3, 5, 99] == test_numbers
        test_numbers = [c.number for c in cp_simple_problem.cells[::-3]]
        assert [99, 3] == test_numbers
        test_numbers = [c.number for c in cp_simple_problem.cells[:6:3]]
        assert [3] == test_numbers
        test_numbers = [c.number for c in cp_simple_problem.cells[5::-1]]
        assert [5, 3, 2, 1] == test_numbers
        test_numbers = [s.number for s in cp_simple_problem.surfaces[1000::10]]
        assert [1000, 1010, 1020] == test_numbers
        test_numbers = [s.number for s in cp_simple_problem.surfaces[:]]
        assert [1000, 1005, 1010, 1015, 1020, 1025] == test_numbers
        test_numbers = [m.number for m in cp_simple_problem.materials[:2]]
        assert [1, 2] == test_numbers
        test_numbers = [m.number for m in cp_simple_problem.materials[::2]]
        assert [2] == test_numbers

    def test_get(self, cp_simple_problem):
        cell_found = cp_simple_problem.cells.get(1)
        assert cp_simple_problem.cells[1] == cell_found
        surf_not_found = cp_simple_problem.surfaces.get(39)  # 39 buried, 0 found
        assert (surf_not_found) is None
        default_mat = cp_simple_problem.materials[3]
        assert cp_simple_problem.materials.get(42, default_mat) == default_mat
        # force a cache miss
        cells = cp_simple_problem.cells
        cells.link_to_problem(None)
        cell = cells[1]
        cell.number = 23
        assert cells.get(23) is cell

    def test_keys(self, cp_simple_problem):
        cell_nums = []
        for c in cp_simple_problem.cells:
            cell_nums.append(c.number)
        cell_keys = []
        for k in cp_simple_problem.cells.keys():
            cell_keys.append(k)
        assert cell_nums == cell_keys
        cells = montepy.Cells()
        # test blank keys
        assert len(list(cells.keys())) == 0

    def test_values(self, cp_simple_problem):
        list_cells = list(cp_simple_problem.cells)
        list_values = list(cp_simple_problem.cells.values())
        assert list_cells == list_values
        cells = montepy.Cells()
        assert len(list(cells.keys())) == 0

    def test_items(self, cp_simple_problem):
        zipped = zip(cp_simple_problem.cells.keys(), cp_simple_problem.cells.values())
        cell_items = cp_simple_problem.cells.items()
        assert tuple(zipped) == tuple(cell_items)
        cells = montepy.Cells()
        assert len(list(cells.keys())) == 0

    def test_eq(_, cp_simple_problem):
        cells = cp_simple_problem.cells
        new_cells = copy.copy(cells)
        assert cells == new_cells
        new_cells = montepy.Cells()
        assert cells != new_cells
        for i in range(len(cells)):
            cell = montepy.Cell()
            cell.number = i + 500
            new_cells.add(cell)
        assert new_cells != cells
        new_cells[501].number = 2
        assert new_cells != cells
        with pytest.raises(TypeError):
            cells == 5

    def test_surface_generators(self, cp_simple_problem):
        answer_num = [1000, 1010]
        spheres = list(cp_simple_problem.surfaces.so)
        assert len(answer_num) == len(spheres)
        for i, sphere in enumerate(spheres):
            assert answer_num[i] == sphere.number

    def test_number_adding_concurancy(self, cp_simple_problem):
        surfaces = copy.deepcopy(cp_simple_problem.surfaces)
        new_surf = copy.deepcopy(surfaces[1005])
        new_surf.number = 5
        surfaces.append(new_surf)
        size = len(surfaces)
        new_surf1 = copy.deepcopy(new_surf)
        with pytest.raises(NumberConflictError):
            surfaces.append(new_surf1)
        surfaces.append_renumber(new_surf1)
        assert len(surfaces) == size + 1
        assert new_surf1.number == 6

    def test_str(self, cp_simple_problem):
        cells = cp_simple_problem.cells
        assert str(cells) == "Cells: [1, 2, 3, 99, 5]"
        key_phrases = [
            "Numbered_object_collection: obj_class: <class 'montepy.cell.Cell'>",
            "Objects: [CELL: 1",
            "Number cache: {1: CELL: 1",
        ]
        for phrase in key_phrases:
            assert phrase in repr(cells)

    def test_data_init(_, cp_simple_problem):
        new_mats = montepy.materials.Materials(
            list(cp_simple_problem.materials), problem=cp_simple_problem
        )
        assert list(new_mats) == list(cp_simple_problem.materials)

    def test_data_append(_, cp_simple_problem):
        prob = cp_simple_problem
        new_mat = copy.deepcopy(next(iter(prob.materials)))
        new_mat.number = prob.materials.request_number()
        prob.materials.append(new_mat)
        assert new_mat in prob.materials
        assert new_mat in prob.data_inputs
        assert prob.data_inputs.count(new_mat) == 1
        # trigger getting data_inputs end
        prob.materials.clear()
        prob.materials.append(new_mat)
        assert new_mat in prob.materials
        assert new_mat in prob.data_inputs
        assert prob.data_inputs.count(new_mat) == 1
        prob.data_inputs.clear()
        prob.materials._last_index = None
        new_mat = copy.deepcopy(next(iter(prob.materials)))
        new_mat.number = prob.materials.request_number()
        prob.materials.append(new_mat)
        assert new_mat in prob.materials
        assert new_mat in prob.data_inputs
        assert prob.data_inputs.count(new_mat) == 1
        # trigger getting index of last material
        prob.materials._last_index = None
        new_mat = copy.deepcopy(next(iter(prob.materials)))
        new_mat.number = prob.materials.request_number()
        prob.materials.append(new_mat)
        assert new_mat in prob.materials
        assert new_mat in prob.data_inputs
        assert prob.data_inputs.count(new_mat) == 1

    def test_data_append_renumber(_, cp_simple_problem):
        prob = cp_simple_problem
        new_mat = copy.deepcopy(next(iter(prob.materials)))
        prob.materials.append_renumber(new_mat)
        assert new_mat in prob.materials
        assert new_mat in prob.data_inputs
        assert prob.data_inputs.count(new_mat) == 1

    def test_data_remove(_, cp_simple_problem):
        prob = cp_simple_problem
        old_mat = next(iter(prob.materials))
        prob.materials.remove(old_mat)
        assert old_mat not in prob.materials
        assert old_mat not in prob.data_inputs
        with pytest.raises(TypeError):
            prob.materials.remove(5)
        mat = montepy.Material()
        with pytest.raises(KeyError):
            prob.materials.remove(mat)
        # do a same number fakeout
        mat = copy.deepcopy(prob.materials[2])
        with pytest.raises(KeyError):
            prob.materials.remove(mat)

    def test_numbered_discard(_, cp_simple_problem):
        mats = cp_simple_problem.materials
        mat = mats[2]
        mats.discard(mat)
        assert mat not in mats
        # no error
        mats.discard(mat)
        mats.discard(5)

    def test_numbered_contains(_, cp_simple_problem):
        mats = cp_simple_problem.materials
        mat = mats[2]
        assert mat in mats
        assert 5 not in mats
        mat = montepy.Material()
        mat.number = 100
        assert mat not in mats
        # num cache fake out
        mat.number = 2
        assert mat not in mats

    @pytest.fixture
    def mats_sets(_):
        mats1 = montepy.Materials()
        mats2 = montepy.Materials()
        for i in range(1, 10):
            mat = montepy.Material()
            mat.number = i
            mats1.append(mat)
        for i in range(5, 15):
            mat = montepy.Material()
            mat.number = i
            mats2.append(mat)
        return (mats1, mats2)

    @pytest.mark.parametrize(
        "name, operator",
        [
            ("and", lambda a, b: a & b),
            ("or", lambda a, b: a | b),
            ("sub", lambda a, b: a - b),
            ("xor", lambda a, b: a ^ b),
            ("sym diff", lambda a, b: a.symmetric_difference(b)),
        ],
    )
    def test_numbered_set_logic(_, mats_sets, name, operator):
        mats1, mats2 = mats_sets
        mats1_nums = set(mats1.keys())
        mats2_nums = set(mats2.keys())
        new_mats = operator(mats1, mats2)
        new_nums = set(new_mats.keys())
        assert new_nums == operator(mats1_nums, mats2_nums)

    @pytest.mark.parametrize(
        "name",
        ["iand", "ior", "isub", "ixor", "sym_diff", "diff", "union", "intersection"],
    )
    def test_numbered_set_logic_update(_, mats_sets, name):
        def operator(a, b):
            if name == "iand":
                a &= b
            elif name == "ior":
                a |= b
            elif name == "isub":
                a -= b
            elif name == "ixor":
                a ^= b
            elif name == "sym_diff":
                a.symmetric_difference_update(b)
            elif name == "diff":
                a.difference_update(b)
            elif name == "union":
                a.update(b)
            elif name == "intersection":
                a.intersection_update(b)

        mats1, mats2 = mats_sets
        mats1_nums = set(mats1.keys())
        mats2_nums = set(mats2.keys())
        operator(mats1, mats2)
        new_nums = set(mats1.keys())
        operator(mats1_nums, mats2_nums)
        assert new_nums == mats1_nums

    @pytest.mark.parametrize(
        "name, operator",
        [
            ("le", lambda a, b: a <= b),
            ("lt", lambda a, b: a < b),
            ("ge", lambda a, b: a >= b),
            ("gt", lambda a, b: a > b),
            ("subset", lambda a, b: a.issubset(b)),
            ("superset", lambda a, b: a.issuperset(b)),
            ("disjoint", lambda a, b: a.isdisjoint(b)),
        ],
    )
    def test_numbered_set_logic_test(_, mats_sets, name, operator):
        mats1, mats2 = mats_sets
        mats1_nums = set(mats1.keys())
        mats2_nums = set(mats2.keys())
        answer = operator(mats1, mats2)
        assert answer == operator(mats1_nums, mats2_nums)

    @pytest.mark.parametrize(
        "name, operator",
        [
            ("intersection", lambda a, *b: a.intersection(*b)),
            ("union", lambda a, *b: a.union(*b)),
            ("difference", lambda a, *b: a.difference(*b)),
        ],
    )
    def test_numbered_set_logic_multi(_, mats_sets, name, operator):
        mats3 = montepy.Materials()
        for i in range(7, 19):
            mat = montepy.Material()
            mat.number = i
            mats3.add(mat)
        mats1, mats2 = mats_sets
        mats1_nums = set(mats1.keys())
        mats2_nums = set(mats2.keys())
        mats3_nums = set(mats3.keys())
        new_mats = operator(mats1, mats2, mats3)
        new_nums = set(new_mats.keys())
        assert new_nums == operator(mats1_nums, mats2_nums, mats3_nums)

    def test_numbered_set_logic_bad(_):
        mats = montepy.Materials()
        with pytest.raises(TypeError):
            mats & 5
        with pytest.raises(TypeError):
            mats &= {5}
        with pytest.raises(TypeError):
            mats |= {5}
        with pytest.raises(TypeError):
            mats -= {5}
        with pytest.raises(TypeError):
            mats ^= {5}
        with pytest.raises(TypeError):
            mats > 5
        with pytest.raises(TypeError):
            mats.union(5)

    def test_data_delete(_, cp_simple_problem):
        prob = cp_simple_problem
        old_mat = next(iter(prob.materials))
        del prob.materials[old_mat.number]
        assert old_mat not in prob.materials
        assert old_mat not in prob.data_inputs
        with pytest.raises(TypeError):
            del prob.materials["foo"]

    def test_data_clear(_, cp_simple_problem):
        data_len = len(cp_simple_problem.data_inputs)
        mat_len = len(cp_simple_problem.materials)
        cp_simple_problem.materials.clear()
        assert len(cp_simple_problem.materials) == 0
        assert len(cp_simple_problem.data_inputs) == data_len - mat_len

    def test_data_pop(_, cp_simple_problem):
        old_mat = next(reversed(list(cp_simple_problem.materials)))
        old_len = len(cp_simple_problem.materials)
        popper = cp_simple_problem.materials.pop()
        assert popper is old_mat
        assert len(cp_simple_problem.materials) == old_len - 1
        assert old_mat not in cp_simple_problem.materials
        assert old_mat not in cp_simple_problem.data_inputs
        with pytest.raises(TypeError):
            cp_simple_problem.materials.pop("foo")

    def test_numbered_starting_number(_):
        cells = montepy.Cells()
        assert cells.starting_number == 1
        cells.starting_number = 5
        assert cells.starting_number == 5
        with pytest.raises(TypeError):
            cells.starting_number = "hi"
        with pytest.raises(ValueError):
            cells.starting_number = -1

    # disable function scoped fixtures
    @settings(suppress_health_check=[hypothesis.HealthCheck.function_scoped_fixture])
    @given(start_num=st.integers(), step=st.integers())
    def test_num_collect_clone(_, read_simple_problem, start_num, step):
        cp_simple_problem = copy.deepcopy(read_simple_problem)
        surfs = copy.deepcopy(cp_simple_problem.surfaces)
        if start_num <= 0 or step <= 0:
            with pytest.raises(ValueError):
                surfs.clone(start_num, step)
            return
        for clear in [False, True]:
            if clear:
                surfs.link_to_problem(None)
            new_surfs = surfs.clone(start_num, step)
            for new_surf, old_surf in zip(new_surfs, surfs):
                assert new_surf is not old_surf
                assert new_surf.surface_type == old_surf.surface_type
                assert new_surf.number != old_surf.number

    @settings(
        suppress_health_check=[hypothesis.HealthCheck.function_scoped_fixture],
        deadline=500,
    )
    @given(
        start_num=st.integers(),
        step=st.integers(),
        clone_mat=st.booleans(),
        clone_region=st.booleans(),
    )
    def test_cells_clone(
        _, read_simple_problem, start_num, step, clone_mat, clone_region
    ):
        cp_simple_problem = copy.deepcopy(read_simple_problem)
        cells = copy.deepcopy(cp_simple_problem.cells)
        if start_num <= 0 or step <= 0:
            with pytest.raises(ValueError):
                cells.clone(starting_number=start_num, step=step)
            return
        for clear in [False, True]:
            if clear:
                cells.link_to_problem(None)
            new_cells = cells.clone(clone_mat, clone_region, start_num, step)
            for new_cell, old_cell in zip(new_cells, cells):
                assert new_cell is not old_cell
                assert new_cell.number != old_cell.number
                assert new_cell.geometry is not old_cell.geometry
                if clone_mat and old_cell.material:
                    assert new_cell.material is not old_cell.material
                else:
                    assert new_cell.material == old_cell.material
                if clone_region:
                    if len(old_cell.surfaces) > 0:
                        assert new_cell.surfaces != old_cell.surfaces
                    if len(old_cell.complements) > 0:
                        assert new_cell.complements != old_cell.complements
                else:
                    assert new_cell.surfaces == old_cell.surfaces
                    assert new_cell.complements == old_cell.complements
                assert new_cell.importance.neutron == old_cell.importance.neutron

    def test_num_collect_clone_default(_, cp_simple_problem):
        surfs = copy.deepcopy(cp_simple_problem.surfaces)
        for clear in [False, True]:
            if clear:
                surfs.link_to_problem(None)
            new_surfs = surfs.clone()
            for new_surf, old_surf in zip(new_surfs, surfs):
                assert new_surf is not old_surf
                assert new_surf.surface_type == old_surf.surface_type
                assert new_surf.number != old_surf.number

    def test_num_collect_link_problem(_, cp_simple_problem):
        cells = montepy.Cells()
        cells.link_to_problem(cp_simple_problem)
        assert cells._problem == cp_simple_problem
        cells.link_to_problem(None)
        assert cells._problem is None
        with pytest.raises(TypeError):
            cells.link_to_problem("hi")

    @pytest.mark.parametrize(
        "args, error",
        [
            (("c", 1), TypeError),
            ((1, "d"), TypeError),
            ((-1, 1), ValueError),
            ((0, 1), ValueError),
            ((1, 0), ValueError),
            ((1, -1), ValueError),
        ],
    )
    def test_num_collect_clone_bad(_, cp_simple_problem, args, error):
        surfs = cp_simple_problem.surfaces
        with pytest.raises(error):
            surfs.clone(*args)


class TestMaterials:

    @pytest.fixture(scope="class")
    def m0_prob(_):
        return montepy.read_input(
            os.path.join("tests", "inputs", "test_importance.imcnp")
        )

    @pytest.fixture
    def cp_m0_prob(_, m0_prob):
        return copy.deepcopy(m0_prob)

    def test_m0_defaults(_, m0_prob):
        prob = m0_prob
        assert prob.materials.default_libraries["nlib"] == "00c"
        assert prob.materials.default_libraries["plib"] == "80p"
        assert prob.materials.default_libraries["alib"] is None

    def test_m0_defaults_fresh(_):
        prob = montepy.MCNP_Problem("")
        prob.materials.default_libraries["nlib"] = "00c"
        prob.materials.default_libraries["plib"] = "80p"
        assert prob.materials.default_libraries["nlib"] == "00c"
        assert prob.materials.default_libraries["plib"] == "80p"
        assert prob.materials.default_libraries["alib"] is None

    @pytest.mark.parametrize(
        "nuclides, threshold, num",
        [
            (("26054", "26056"), 1.0, 1),
            ((montepy.Nuclide("H-1"),), 0.0, 1),
            (("B",), 1.0, 0),
        ],
    )
    def test_get_containing_all(_, m0_prob, nuclides, threshold, num):
        ret = list(m0_prob.materials.get_containing_all(*nuclides, threshold=threshold))
        assert len(ret) == num
        for mat in ret:
            assert isinstance(mat, montepy.Material)
        with pytest.raises(TypeError):
            next(m0_prob.materials.get_containing_all(m0_prob))

    @pytest.mark.parametrize(
        "nuclides, threshold, num",
        [
            (("26054", "26056"), 1.0, 1),
            ((montepy.Nuclide("H-1"),), 0.0, 1),
            (("B",), 1.0, 0),
            (("U-235", "H-1"), 0.0, 2),
        ],
    )
    def test_get_containing_any(_, m0_prob, nuclides, threshold, num):
        ret = list(m0_prob.materials.get_containing_any(*nuclides, threshold=threshold))
        assert len(ret) == num
        for mat in ret:
            assert isinstance(mat, montepy.Material)
        with pytest.raises(TypeError):
            next(m0_prob.materials.get_containing_any(m0_prob))

    @pytest.fixture
    def h2o(_):
        mat = montepy.Material()
        mat.number = 1
        mat.add_nuclide("H-1.80c", 2.0)
        mat.add_nuclide("O-16.80c", 1.0)
        return mat

    @pytest.fixture
    def mass_h2o(_):
        mat = montepy.Material()
        mat.number = 1
        mat.is_atom_fraction = False
        mat.add_nuclide("H-1.80c", 2.0)
        mat.add_nuclide("O-16.80c", 1.0)
        return mat

    @pytest.fixture
    def boric_acid(_):
        mat = montepy.Material()
        mat.number = 2
        for nuclide, fraction in {
            "1001.80c": 3.0,
            "B-10.80c": 1.0 * 0.189,
            "B-11.80c": 1.0 * 0.796,
            "O-16.80c": 3.0,
        }.items():
            mat.add_nuclide(nuclide, fraction)
        return mat

    @pytest.fixture
    def mats_dict(_, h2o, mass_h2o, boric_acid):
        return {"h2o": h2o, "mass_h2o": mass_h2o, "boric_acid": boric_acid}

    @pytest.mark.parametrize(
        "args, error, use_fixture",
        [
            (("hi", [1]), TypeError, False),
            ((["hi"], [1]), TypeError, False),
            (([], [1]), ValueError, False),  # empty materials
            ((["h2o", "mass_h2o"], [1, 2]), ValueError, True),  # mismatch is_atom
            ((["h2o", "boric_acid"], [1.0]), ValueError, True),  # mismatch lengths
            ((["h2o", "boric_acid"], "hi"), TypeError, True),
            ((["h2o", "boric_acid"], ["hi"]), TypeError, True),
            ((["h2o", "boric_acid"], [-1.0, 2.0]), ValueError, True),
            ((["h2o", "boric_acid"], [1.0, 2.0], "hi"), TypeError, True),
            ((["h2o", "boric_acid"], [1.0, 2.0], -1), ValueError, True),
            ((["h2o", "boric_acid"], [1.0, 2.0], 1, "hi"), TypeError, True),
            ((["h2o", "boric_acid"], [1.0, 2.0], 1, -1), ValueError, True),
        ],
    )
    def test_mix_bad(_, mats_dict, args, error, use_fixture):
        if use_fixture:
            mats = []
            for mat in args[0]:
                mats.append(mats_dict[mat])
            args = (mats,) + args[1:]
        with pytest.raises(error):
            mats = montepy.Materials()
            mats.mix(*args)

    @given(
        starting_num=st.one_of(st.none(), st.integers(1)),
        step=st.one_of(st.none(), st.integers(1)),
    )
    def test_mix(_, starting_num, step):
        mat = montepy.Material()
        mat.number = 1
        mat.add_nuclide("H-1.80c", 2.0)
        mat.add_nuclide("O-16.80c", 1.0)
        parents = [mat]
        mat = montepy.Material()
        mat.number = 2
        for nuclide, fraction in {
            "1001.80c": 3.0,
            "B-10.80c": 1.0 * 0.189,
            "B-11.80c": 1.0 * 0.796,
            "O-16.80c": 3.0,
        }.items():
            mat.add_nuclide(nuclide, fraction)
        parents.append(mat)
        boron_conc = 10 * 1e-6
        fractions = [1 - boron_conc, boron_conc]
        mats = montepy.Materials()
        for par in parents:
            mats.append(par)
        new_mat = mats.mix(
            parents,
            fractions,
            starting_num,
            step,
        )
        assert sum(new_mat.values) == pytest.approx(
            1.0
        )  # should normalize to 1 with fractions
        assert new_mat.is_atom_fraction == parents[0].is_atom_fraction
        flat_fracs = []
        for par, frac in zip(parents, fractions):
            par.normalize()
            flat_fracs += [frac] * len(par)
        for (new_nuc, new_frac), (old_nuc, old_frac), fraction in zip(
            new_mat, it.chain(*parents), flat_fracs
        ):
            assert new_nuc == old_nuc
            assert new_nuc is not old_nuc
            assert new_frac == pytest.approx(old_frac * fraction)
        if starting_num is None:
            starting_num = mats.starting_number
        if step is None:
            step = mats.step
        if starting_num not in [p.number for p in parents]:
            assert new_mat.number == starting_num
        else:
            assert (new_mat.number - starting_num) % step == 0



# Copyright 2024-2025, Battelle Energy Alliance, LLC All Rights Reserved.
from __future__ import annotations
from abc import ABC, ABCMeta, abstractmethod
import copy
import functools
import itertools as it
import sys
import textwrap
from typing import Union
import warnings
import weakref

from montepy.errors import *
from montepy.constants import (
    BLANK_SPACE_CONTINUE,
    COMMENT_FINDER,
    get_max_line_length,
    rel_tol,
    abs_tol,
)
from montepy.input_parser.syntax_node import (
    CommentNode,
    PaddingNode,
    ParametersNode,
    ValueNode,
)
import montepy

InitInput = Union[montepy.input_parser.mcnp_input.Input, str]


class _ExceptionContextAdder(ABCMeta):
    """A metaclass for wrapping all class properties and methods in :func:`~montepy.errors.add_line_number_to_exception`."""

    @staticmethod
    def _wrap_attr_call(func):
        """Wraps the function, and returns the modified function."""

        @functools.wraps(func)
        def wrapped(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as e:
                if len(args) > 0 and isinstance(args[0], MCNP_Object):
                    self = args[0]
                    if hasattr(self, "_handling_exception"):
                        raise e
                    self._handling_exception = True
                    try:
                        add_line_number_to_exception(e, self)
                    finally:
                        del self._handling_exception
                else:
                    raise e

        if isinstance(func, staticmethod):
            return staticmethod(wrapped)
        if isinstance(func, classmethod):
            return classmethod(wrapped)
        return wrapped

    def __new__(meta, classname, bases, attributes):
        """This will replace all properties and callable attributes with
        wrapped versions.
        """
        new_attrs = {}
        for key, value in attributes.items():
            if key.startswith("_"):
                new_attrs[key] = value
            if callable(value):
                new_attrs[key] = _ExceptionContextAdder._wrap_attr_call(value)
            elif isinstance(value, property):
                new_props = {}
                for attr_name in {"fget", "fset", "fdel", "doc"}:
                    try:
                        assert getattr(value, attr_name)
                        new_props[attr_name] = _ExceptionContextAdder._wrap_attr_call(
                            getattr(value, attr_name)
                        )
                    except (AttributeError, AssertionError):
                        new_props[attr_name] = None

                new_attrs[key] = property(**new_props)
            else:
                new_attrs[key] = value
        cls = super().__new__(meta, classname, bases, new_attrs)
        return cls


class MCNP_Object(ABC, metaclass=_ExceptionContextAdder):
    """Abstract class for semantic representations of MCNP inputs.

    Parameters
    ----------
    input : Union[Input, str]
        The Input syntax object this will wrap and parse.
    parser : MCNP_Parser
        The parser object to parse the input with.
    """

    def __init__(
        self,
        input: InitInput,
        parser: montepy.input_parser.parser_base.MCNP_Parser,
    ):
        try:
            self._BLOCK_TYPE
        except AttributeError:
            self._BLOCK_TYPE = montepy.input_parser.block_type.BlockType.DATA
        self._problem_ref = None
        self._parameters = ParametersNode()
        self._input = None
        if input:
            if not isinstance(input, (montepy.input_parser.mcnp_input.Input, str)):
                raise TypeError(f"input must be an Input or str. {input} given.")
            if isinstance(input, str):
                input = montepy.input_parser.mcnp_input.Input(
                    input.split("\n"), self._BLOCK_TYPE
                )
            try:
                try:
                    parser.restart()
                # raised if restarted without ever parsing
                except AttributeError as e:
                    pass
                self._tree = parser.parse(input.tokenize(), input)
                self._input = input
            except ValueError as e:
                raise MalformedInputError(
                    input, f"Error parsing object of type: {type(self)}: {e.args[0]}"
                ).with_traceback(e.__traceback__)
            if self._tree is None:
                raise ParsingError(
                    input,
                    "",
                    parser.log.clear_queue(),
                )
            if "parameters" in self._tree:
                self._parameters = self._tree["parameters"]

    def __setattr__(self, key, value):
        # handle properties first
        if hasattr(type(self), key):
            descriptor = getattr(type(self), key)
            if isinstance(descriptor, property):
                descriptor.__set__(self, value)
                return
        # handle _private second
        if key.startswith("_"):
            super().__setattr__(key, value)
        else:
            # kwargs added in 3.10
            if sys.version_info >= (3, 10):
                raise AttributeError(
                    f"'{type(self).__name__}' object has no attribute '{key}'",
                    obj=self,
                    name=key,
                )
            raise AttributeError(
                f"'{type(self).__name__}' object has no attribute '{key}'",
            )

    @staticmethod
    def _generate_default_node(value_type: type, default, padding: str = " "):
        """Generates a "default" or blank ValueNode.

        None is generally a safe default value to provide.

        Parameters
        ----------
        value_type : Class
            the data type for the ValueNode.
        default : value_type
            the default value to provide (type needs to agree with
            value_type)
        padding : str, None
            the string to provide to the PaddingNode. If None no
            PaddingNode will be added.

        Returns
        -------
        ValueNode
            a new ValueNode with the requested information.
        """
        if padding:
            padding_node = PaddingNode(padding)
        else:
            padding_node = None
        if default is None or isinstance(default, montepy.input_parser.mcnp_input.Jump):
            return ValueNode(default, value_type, padding_node)
        return ValueNode(str(default), value_type, padding_node)

    @property
    def parameters(self) -> dict[str, str]:
        """A dictionary of the additional parameters for the object.

        e.g.: ``1 0 -1 u=1 imp:n=0.5`` has the parameters
        ``{"U": "1", "IMP:N": "0.5"}``

        Returns
        -------
        unknown
            a dictionary of the key-value pairs of the parameters.


        :rytpe: dict
        """
        return self._parameters

    @abstractmethod
    def _update_values(self):
        """Method to update values in syntax tree with new values.

        Generally when :func:`~montepy.utilities.make_prop_val_node` this is not necessary to do,
        but when :func:`~montepy.utilities.make_prop_pointer` is used it is necessary.
        The most common need is to update a value based on the number for an object pointed at,
        e.g., the material number in a cell definition.

        """
        pass

    def format_for_mcnp_input(self, mcnp_version: tuple[int]) -> list[str]:
        """Creates a string representation of this MCNP_Object that can be
        written to file.

        Parameters
        ----------
        mcnp_version : tuple
            The tuple for the MCNP version that must be exported to.

        Returns
        -------
        list
            a list of strings for the lines that this input will occupy.
        """
        self.validate()
        self._update_values()
        self._tree.check_for_graveyard_comments()
        message = None
        with warnings.catch_warnings(record=True) as ws:
            lines = self.wrap_string_for_mcnp(self._tree.format(), mcnp_version, True)
        self._flush_line_expansion_warning(lines, ws)
        return lines

    def _flush_line_expansion_warning(self, lines, ws):
        if not ws:
            return
        message = f"""The input had a value expand that may change formatting.
The new input was:\n\n"""
        for line in lines:
            message += f"    {line}"
        width = 15
        message += f"\n\n    {'old value': ^{width}s} {'new value': ^{width}s}"
        message += f"\n    {'':-^{width}s} {'':-^{width}s}\n"
        olds = []
        news = []
        for w in ws:
            warning = w.message
            formatter = f"    {{w.og_value: >{width}}} {{w.new_value: >{width}}}\n"
            message += formatter.format(w=warning)
            olds.append(warning.og_value)
            news.append(warning.new_value)
        if message is not None:
            warning = LineExpansionWarning(message)
            warning.olds = olds
            warning.news = news
            warnings.warn(warning, stacklevel=4)

    @property
    def comments(self) -> list[PaddingNode]:
        """The comments associated with this input if any.

        This includes all ``C`` comments before this card that aren't part of another card,
        and any comments that are inside this card.

        Returns
        -------
        list
            a list of the comments associated with this comment.
        """
        return list(self._tree.comments)

    @property
    def leading_comments(self) -> list[PaddingNode]:
        """Any comments that come before the beginning of the input proper.

        Returns
        -------
        list
            the leading comments.
        """
        return list(self._tree["start_pad"].comments)

    @leading_comments.setter
    def leading_comments(self, comments):
        if not isinstance(comments, (list, tuple, CommentNode)):
            raise TypeError(
                f"Comments must be a CommentNode, or a list of Comments. {comments} given."
            )
        if isinstance(comments, CommentNode):
            comments = [comments]
        if isinstance(comments, (list, tuple)):
            for comment in comments:
                if not isinstance(comment, CommentNode):
                    raise TypeError(
                        f"Comments must be a CommentNode, or a list of Comments. {comment} given."
                    )

        for i, comment in enumerate(comments):
            if not isinstance(comment, CommentNode):
                raise TypeError(
                    f"Comment must be a CommentNode. {comment} given at index {i}."
                )
        new_nodes = list(*zip(comments, it.cycle(["\n"])))
        if self._tree["start_pad"] is None:
            self._tree["start_pad"] = PaddingNode(" ")
        self._tree["start_pad"]._nodes = new_nodes

    @leading_comments.deleter
    def leading_comments(self):
        self._tree["start_pad"]._delete_trailing_comment()

    @staticmethod
    def wrap_string_for_mcnp(
        string, mcnp_version, is_first_line, suppress_blank_end=True
    ) -> list[str]:
        """Wraps the list of the words to be a well formed MCNP input.

        multi-line inputs will be handled by using the indentation format,
        and not the "&" method.

        Parameters
        ----------
        string : str
            A long string with new lines in it, that needs to be chunked
            appropriately for MCNP inputs
        mcnp_version : tuple
            the tuple for the MCNP that must be formatted for.
        is_first_line : bool
            If true this will be the beginning of an MCNP input. The
            first line will not be indented.
        suppress_blank_end : bool
            Whether or not to suppress any blank lines that would be
            added to the end. Good for anywhere but cell modifiers in
            the cell block.

        Returns
        -------
        list
            A list of strings that can be written to an input file, one
            item to a line.
        """
        line_length = get_max_line_length(mcnp_version)
        indent_length = BLANK_SPACE_CONTINUE
        strings = string.splitlines()
        if is_first_line:
            initial_indent = 0
        else:
            initial_indent = indent_length
        wrapper = textwrap.TextWrapper(
            width=line_length,
            initial_indent=" " * initial_indent,
            subsequent_indent=" " * indent_length,
            drop_whitespace=False,
        )
        ret = []
        for line in strings:
            buffer = wrapper.wrap(line)
            if len(buffer) > 1:
                # don't warn for comments, nor line wrap
                # this order assumes that comment overruns are rare
                if COMMENT_FINDER.match(line):
                    buffer = [line]
                elif "$" in line:
                    parts = line.split("$")
                    buffer = wrapper.wrap(parts[0])
                    buffer[-1] = "$".join([buffer[-1]] + parts[1:])
                else:
                    warning = LineExpansionWarning(
                        f"The line exceeded the maximum length allowed by MCNP, and was split. The line was:\n{line}"
                    )
                    warning.cause = "line"
                    warning.og_value = line
                    warning.new_value = buffer
                    warnings.warn(
                        warning,
                        LineExpansionWarning,
                        stacklevel=2,
                    )
            # lazy final guard against extra lines
            if suppress_blank_end:
                buffer = [s for s in buffer if s.strip()]
            ret += buffer
        return ret

    def validate(self):
        """Validates that the object is in a usable state."""
        pass

    def link_to_problem(self, problem: montepy.mcnp_problem.MCNP_Problem):
        """Links the input to the parent problem for this input.

        This is done so that inputs can find links to other objects.

        Parameters
        ----------
        problem : MCNP_Problem
            The problem to link this input to.
        """
        if not isinstance(problem, (montepy.mcnp_problem.MCNP_Problem, type(None))):
            raise TypeError("problem must be an MCNP_Problem")
        if problem is None:
            self._problem_ref = None
        else:
            self._problem_ref = weakref.ref(problem)

    @property
    def _problem(self) -> montepy.MCNP_Problem:
        if self._problem_ref is not None:
            return self._problem_ref()
        return None

    @_problem.setter
    def _problem(self, problem):
        if problem is None:
            self._problem_ref = None
            return
        self.link_to_problem(problem)

    @property
    def trailing_comment(self) -> list[PaddingNode]:
        """The trailing comments and padding of an input.

        Generally this will be blank as these will be moved to be a leading comment for the next input.

        Returns
        -------
        list
            the trailing ``c`` style comments and intermixed padding
            (e.g., new lines)
        """
        return self._tree.get_trailing_comment()

    def _delete_trailing_comment(self):
        self._tree._delete_trailing_comment()

    def _grab_beginning_comment(self, padding: list[PaddingNode], last_obj=None):
        if padding:
            self._tree["start_pad"]._grab_beginning_comment(padding)

    def __getstate__(self):
        state = self.__dict__.copy()
        bad_keys = {"_problem_ref", "_parser"}
        for key in bad_keys:
            if key in state:
                del state[key]
        return state

    def __setstate__(self, crunchy_data):
        crunchy_data["_problem_ref"] = None
        self.__dict__.update(crunchy_data)

    def clone(self) -> montepy.mcnp_object.MCNP_Object:
        """Create a new independent instance of this object.

        Returns
        -------
        type(self)
            a new instance identical to this object.
        """
        return copy.deepcopy(self)



# Copyright 2024-2025, Battelle Energy Alliance, LLC All Rights Reserved.
from __future__ import annotations
from abc import ABC
import itertools as it
import typing
import weakref
from numbers import Integral

import montepy
from montepy.numbered_mcnp_object import Numbered_MCNP_Object
from montepy.errors import *
from montepy.utilities import *


def _enforce_positive(self, num):
    if num <= 0:
        raise ValueError(f"Value must be greater than 0. {num} given.")


class NumberedObjectCollection(ABC):
    """A collections of MCNP objects.

    .. _collect ex:

    Examples
    ________

    Accessing Objects
    ^^^^^^^^^^^^^^^^^

    It quacks like a dict, it acts like a dict, but it's a list.

    The items in the collection are accessible by their number.
    For instance to get the Cell with a number of 2 you can just say:

    .. doctest:: python

        >>> import montepy
        >>> problem = montepy.read_input("tests/inputs/test.imcnp")
        >>> cell = problem.cells[2]
        >>> print(cell)
        CELL: 2, mat: 2, DENS: 8.0 atom/b-cm

    You can also add, and delete items like you would in a dictionary normally.
    Though :func:`append` and :func:`add` are the preferred way of adding items.
    When adding items by key the key given is actually ignored.

    .. testcode::

        import montepy
        problem = montepy.read_input("tests/inputs/test.imcnp")
        cell = montepy.Cell()
        cell.number = 25
        # this will actually append ignoring the key given
        problem.cells[3] = cell
        print(problem.cells[3] is cell)
        del problem.cells[25]
        print(cell not in problem.cells)

    This shows:

    .. testoutput::

        False
        True

    Slicing a Collection
    ^^^^^^^^^^^^^^^^^^^^

    Unlike dictionaries this collection also supports slices e.g., ``[1:3]``.
    This will return a new :class:`NumberedObjectCollection` with objects
    that have numbers that fit that slice.

    .. testcode::

        for cell in problem.cells[1:3]:
            print(cell.number)

    Which shows

    .. testoutput::

        1
        2
        3

    Because MCNP numbered objects start at 1, so do the indices.
    The slices are effectively 1-based and endpoint-inclusive.
    This means rather than the normal behavior of [0:5] excluding the index
    5, 5 would be included.

    Set-Like Operations
    ^^^^^^^^^^^^^^^^^^^

    .. versionchanged:: 1.0.0

         Introduced set-like behavior.

    These collections act like `sets <https://docs.python.org/3/library/stdtypes.html#set>`_.
    The supported operators are: ``&``, ``|``, ``-``, ``^``, ``<``, ``<=``, ``>``, ``>=``, ``==``.
    See the set documentation for how these operators function.
    The set operations are applied to the object numbers.
    The corresponding objects are then taken to form a new instance of this collection.
    The if both collections have objects with the same number but different objects,
    the left-hand-side's object is taken.

    .. testcode::

        cells1 = montepy.Cells()

        for i in range(5, 10):
            cell = montepy.Cell()
            cell.number = i
            cells1.add(cell)

        cells2 = montepy.Cells()

        for i in range(8, 15):
            cell = montepy.Cell()
            cell.number = i
            cells2.add(cell)

        overlap = cells1 & cells2

        # The only overlapping numbers are 8, 9, 10

        print({8, 9} == set(overlap.keys()))

    This would print:

    .. testoutput::

        True

    Other set-like functions are: :func:`difference`, :func:`difference_update`,
    :func:`intersection`, :func:`isdisjoint`, :func:`issubset`, :func:`issuperset`,
    :func:`symmetric_difference`, :func:`symmetric_difference_update`, :func:`union`, :func:`discard`, and :func:`update`.

    Parameters
    ----------
    obj_class : type
        the class of numbered objects being collected
    objects : list
        the list of cells to start with if needed
    problem : MCNP_Problem
        the problem to link this collection to.
    """

    def __init__(
        self,
        obj_class: type,
        objects: list = None,
        problem: montepy.MCNP_Problem = None,
    ):
        self.__num_cache = {}
        assert issubclass(obj_class, Numbered_MCNP_Object)
        self._obj_class = obj_class
        self._objects = []
        self._start_num = 1
        self._step = 1
        self._problem_ref = None
        if problem is not None:
            self._problem_ref = weakref.ref(problem)
        if objects:
            if not isinstance(objects, list):
                raise TypeError("NumberedObjectCollection must be built from a list")
            for obj in objects:
                if not isinstance(obj, obj_class):
                    raise TypeError(
                        f"The object: {obj} being added to a NumberedObjectCollection is not of type {obj_class}"
                    )
                if obj.number in self.__num_cache:
                    raise NumberConflictError(
                        (
                            f"When building {self} there was a numbering conflict between: "
                            f"{obj} and {self[obj.number]}"
                        )
                    )
                self.__num_cache[obj.number] = obj
            self._objects = objects

    def link_to_problem(self, problem):
        """Links the card to the parent problem for this card.

        This is done so that cards can find links to other objects.

        Parameters
        ----------
        problem : MCNP_Problem
            The problem to link this card to.
        """
        if not isinstance(problem, (montepy.mcnp_problem.MCNP_Problem, type(None))):
            raise TypeError("problem must be an MCNP_Problem")
        if problem is None:
            self._problem_ref = None
        else:
            self._problem_ref = weakref.ref(problem)
        for obj in self:
            obj.link_to_problem(problem)

    @property
    def _problem(self):
        if self._problem_ref is not None:
            return self._problem_ref()
        return None

    def __getstate__(self):
        state = self.__dict__.copy()
        weakref_key = "_problem_ref"
        if weakref_key in state:
            del state[weakref_key]
        return state

    def __setstate__(self, crunchy_data):
        crunchy_data["_problem_ref"] = None
        self.__dict__.update(crunchy_data)

    @property
    def numbers(self):
        """A generator of the numbers being used.

        Returns
        -------
        generator
        """
        for obj in self._objects:
            # update cache every time we go through all objects
            self.__num_cache[obj.number] = obj
            yield obj.number

    def check_number(self, number):
        """Checks if the number is already in use, and if so raises an error.

        Parameters
        ----------
        number : int
            The number to check.

        Raises
        ------
        NumberConflictError
            if this number is in use.
        """
        if not isinstance(number, Integral):
            raise TypeError("The number must be an int")
        conflict = False
        # only can trust cache if being
        if self._problem:
            if number in self.__num_cache:
                conflict = True
        else:
            if number in self.numbers:
                conflict = True
        if conflict:
            raise NumberConflictError(
                f"Number {number} is already in use for the collection: {type(self).__name__} by {self[number]}"
            )

    def _update_number(self, old_num, new_num, obj):
        """Updates the number associated with a specific object in the internal cache.

        Parameters
        ----------
        old_num : int
            the previous number the object had.
        new_num : int
            the number that is being set to.
        obj : self._obj_class
            the object being updated.
        """
        # don't update numbers you don't own
        if self.__num_cache.get(old_num, None) is not obj:
            return
        self.__num_cache.pop(old_num, None)
        self.__num_cache[new_num] = obj

    @property
    def objects(self):
        """Returns a shallow copy of the internal objects list.

        The list object is a new instance, but the underlying objects
        are the same.

        Returns
        -------
        list
        """
        return self._objects[:]

    def pop(self, pos=-1):
        """Pop the final items off of the collection

        Parameters
        ----------
        pos : int
            The index of the element to pop from the internal list.

        Returns
        -------
        Numbered_MCNP_Object
            the final elements
        """
        if not isinstance(pos, Integral):
            raise TypeError("The index for popping must be an int")
        obj = self._objects[pos]
        self.__internal_delete(obj)
        return obj

    def clear(self):
        """Removes all objects from this collection."""
        self._objects.clear()
        self.__num_cache.clear()

    def extend(self, other_list):
        """Extends this collection with another list.

        Parameters
        ----------
        other_list : list
            the list of objects to add.

        Raises
        ------
        NumberConflictError
            if these items conflict with existing elements.
        """
        if not isinstance(other_list, (list, type(self))):
            raise TypeError("The extending list must be a list")
        # this is the optimized version to get all numbers
        if self._problem:
            nums = set(self.__num_cache)
        else:
            nums = set(self.numbers)
        for obj in other_list:
            if not isinstance(obj, self._obj_class):
                raise TypeError(
                    "The object in the list {obj} is not of type: {self._obj_class}"
                )
            if obj.number in nums:
                raise NumberConflictError(
                    (
                        f"When adding to {type(self).__name__} there was a number collision due to "
                        f"adding {obj} which conflicts with {self[obj.number]}"
                    )
                )
            nums.add(obj.number)
        for obj in other_list:
            self.__internal_append(obj)

    def remove(self, delete):
        """Removes the given object from the collection.

        Parameters
        ----------
        delete : Numbered_MCNP_Object
            the object to delete
        """
        if not isinstance(delete, self._obj_class):
            raise TypeError("")
        candidate = self[delete.number]
        if delete is candidate:
            del self[delete.number]
        else:
            raise KeyError(f"This object is not in this collection")

    def clone(self, starting_number=None, step=None):
        """Create a new instance of this collection, with all new independent
        objects with new numbers.

        This relies mostly on ``copy.deepcopy``.

        Notes
        -----
        If starting_number, or step are not specified :func:`starting_number`,
        and :func:`step` are used as default values.


        .. versionadded:: 0.5.0

        Parameters
        ----------
        starting_number : int
            The starting number to request for a new object numbers.
        step : int
            the step size to use to find a new valid number.

        Returns
        -------
        type(self)
            a cloned copy of this object.
        """
        if not isinstance(starting_number, (Integral, type(None))):
            raise TypeError(
                f"Starting_number must be an int. {type(starting_number)} given."
            )
        if not isinstance(step, (Integral, type(None))):
            raise TypeError(f"step must be an int. {type(step)} given.")
        if starting_number is not None and starting_number <= 0:
            raise ValueError(f"starting_number must be >= 1. {starting_number} given.")
        if step is not None and step <= 0:
            raise ValueError(f"step must be >= 1. {step} given.")
        if starting_number is None:
            starting_number = self.starting_number
        if step is None:
            step = self.step
        objs = []
        for obj in self:
            new_obj = obj.clone(starting_number, step)
            starting_number = new_obj.number
            objs.append(new_obj)
            starting_number = new_obj.number + step
        return type(self)(objs)

    @make_prop_pointer("_start_num", Integral, validator=_enforce_positive)
    def starting_number(self):
        """The starting number to use when an object is cloned.

        Returns
        -------
        int
            the starting number
        """
        pass

    @make_prop_pointer("_step", Integral, validator=_enforce_positive)
    def step(self):
        """The step size to use to find a valid number during cloning.

        Returns
        -------
        int
            the step size
        """
        pass

    def __iter__(self):
        self._iter = self._objects.__iter__()
        return self._iter

    def __str__(self):
        base_class_name = self.__class__.__name__
        numbers = list(self.numbers)
        return f"{base_class_name}: {numbers}"

    def __repr__(self):
        return (
            f"Numbered_object_collection: obj_class: {self._obj_class}, problem: {self._problem}\n"
            f"Objects: {self._objects}\n"
            f"Number cache: {self.__num_cache}"
        )

    def _append_hook(self, obj, initial_load=False):
        """A hook that is called every time append is called."""
        if initial_load:
            return
        if self._problem:
            obj._add_children_objs(self._problem)

    def _delete_hook(self, obj, **kwargs):
        """A hook that is called every time delete is called."""
        pass

    def __internal_append(self, obj, **kwargs):
        """The internal append method.

        This should always be called rather than manually added.

        Parameters
        ----------
        obj
            the obj to append
        **kwargs
            keyword arguments passed through to the append_hook
        """
        if not isinstance(obj, self._obj_class):
            raise TypeError(
                f"Object must be of type: {self._obj_class.__name__}. {obj} given."
            )
        if obj.number in self.__num_cache:
            try:
                if obj is self[obj.number]:
                    return
            # if cache is bad and it's not actually in use ignore it
            except KeyError as e:
                pass
            else:
                raise NumberConflictError(
                    f"Number {obj.number} is already in use for the collection: {type(self).__name__} by {self[obj.number]}"
                )
        self.__num_cache[obj.number] = obj
        self._objects.append(obj)
        self._append_hook(obj, **kwargs)
        if self._problem:
            obj.link_to_problem(self._problem)

    def __internal_delete(self, obj, **kwargs):
        """The internal delete method.

        This should always be called rather than manually added.
        """
        self.__num_cache.pop(obj.number, None)
        self._objects.remove(obj)
        self._delete_hook(obj, **kwargs)

    def add(self, obj: Numbered_MCNP_Object):
        """Add the given object to this collection.

        Parameters
        ----------
        obj : Numbered_MCNP_Object
            The object to add.

        Raises
        ------
        TypeError
            if the object is of the wrong type.
        NumberConflictError
            if this object's number is already in use in the collection.
        """
        self.__internal_append(obj)

    def update(self, *objs: typing.Self):
        """Add the given objects to this collection.

        Notes
        -----

        This is not a thread-safe method.


        .. versionchanged:: 1.0.0

            Changed to be more set like. Accepts multiple arguments. If there is a number conflict,
            the current object will be kept.

        Parameters
        ----------
        *objs : list[Numbered_MCNP_Object]
            The objects to add.

        Raises
        ------
        TypeError
            if the object is of the wrong type.
        NumberConflictError
            if this object's number is already in use in the collection.
        """
        try:
            iter(objs)
        except TypeError:
            raise TypeError(f"Objs must be an iterable. {objs} given.")
        others = []
        for obj in objs:
            if isinstance(obj, list):
                others.append(type(self)(obj))
            else:
                others.append(obj)
        if len(others) == 1:
            self |= others[0]
        else:
            other = others[0].union(*others[1:])
            self |= others

    def append(self, obj, **kwargs):
        """Appends the given object to the end of this collection.

        Parameters
        ----------
        obj : Numbered_MCNP_Object
            the object to add.
        **kwargs
            extra arguments that are used internally.

        Raises
        ------
        NumberConflictError
            if this object has a number that is already in use.
        """
        if not isinstance(obj, self._obj_class):
            raise TypeError(f"object being appended must be of type: {self._obj_class}")
        self.__internal_append(obj, **kwargs)

    def append_renumber(self, obj, step=1):
        """Appends the object, but will renumber the object if collision occurs.

        This behaves like append, except if there is a number collision the object will
        be renumbered to an available number. The number will be incremented by step
        until an available number is found.

        Parameters
        ----------
        obj : Numbered_MCNP_Object
            The MCNP object being added to the collection.
        step : int
            the incrementing step to use to find a new number.

        Returns
        -------
        int
            the number for the object.
        """
        if not isinstance(obj, self._obj_class):
            raise TypeError(f"object being appended must be of type: {self._obj_class}")
        if not isinstance(step, Integral):
            raise TypeError("The step number must be an int")
        number = obj.number
        if self._problem:
            obj.link_to_problem(self._problem)
        try:
            self.append(obj)
        except NumberConflictError:
            number = self.request_number(number, step)
            obj.number = number
            self.append(obj)

        return number

    def request_number(self, start_num=None, step=None):
        """Requests a new available number.

        This method does not "reserve" this number. Objects
        should be immediately added to avoid possible collisions
        caused by shifting numbers of other objects in the collection.

        Notes
        -----
        If starting_number, or step are not specified :func:`starting_number`,
        and :func:`step` are used as default values.


        .. versionchanged:: 0.5.0
            In 0.5.0 the default values were changed to reference :func:`starting_number` and :func:`step`.

        Parameters
        ----------
        start_num : int
            the starting number to check.
        step : int
            the increment to jump by to find new numbers.

        Returns
        -------
        int
            an available number
        """
        if not isinstance(start_num, (Integral, type(None))):
            raise TypeError("start_num must be an int")
        if not isinstance(step, (Integral, type(None))):
            raise TypeError("step must be an int")
        if start_num is None:
            start_num = self.starting_number
        if step is None:
            step = self.step
        number = start_num
        while True:
            try:
                self.check_number(number)
                break
            except NumberConflictError:
                number += step
        return number

    def next_number(self, step=1):
        """Get the next available number, based on the maximum number.

        This works by finding the current maximum number, and then adding the
        stepsize to it.

        Parameters
        ----------
        step : int
            how much to increase the last number by
        """
        if not isinstance(step, Integral):
            raise TypeError("step must be an int")
        if step <= 0:
            raise ValueError("step must be > 0")
        return max(self.numbers) + step

    def __get_slice(self, i: slice):
        """Get a new NumberedObjectCollection over a slice of numbers

        This method implements usage like:
            >>> NumberedObjectCollection[1:49:12]
            [1, 13, 25, 37, 49]

        Any ``slice`` object may be passed.
        The indices are the object numbers.
        Because MCNP numbered objects start at 1, so do the indices.
        They are effectively 1-based and endpoint-inclusive.

        Returns
        -------
        NumberedObjectCollection
        """
        rstep = i.step if i.step is not None else 1
        rstart = i.start
        rstop = i.stop
        if rstep < 0:  # Backwards
            if rstart is None:
                rstart = max(self.numbers)
            if rstop is None:
                rstop = min(self.numbers)
            rstop -= 1
        else:  # Forwards
            if rstart is None:
                rstart = 0
            if rstop is None:
                rstop = max(self.numbers)
            rstop += 1
        numbered_objects = []
        for num in range(rstart, rstop, rstep):
            obj = self.get(num)
            if obj is not None:
                numbered_objects.append(obj)
        # obj_class is always implemented in child classes.
        return type(self)(numbered_objects)

    def __getitem__(self, i):
        if isinstance(i, slice):
            return self.__get_slice(i)
        elif not isinstance(i, Integral):
            raise TypeError("index must be an int or slice")
        ret = self.get(i)
        if ret is None:
            raise KeyError(f"Object with number {i} not found in {type(self)}")
        return ret

    def __delitem__(self, idx):
        if not isinstance(idx, Integral):
            raise TypeError("index must be an int")
        obj = self[idx]
        self.__internal_delete(obj)

    def __setitem__(self, key, newvalue):
        if not isinstance(key, Integral):
            raise TypeError("index must be an int")
        self.append(newvalue)

    def __len__(self):
        return len(self._objects)

    def __iadd__(self, other):
        self.extend(other)
        return self

    def __contains__(self, other):
        if not isinstance(other, self._obj_class):
            return False
        # if cache can be trusted from #563
        if self._problem:
            try:
                if other is self[other.number]:
                    return True
                return False
            except KeyError:
                return False
        return other in self._objects

    def __set_logic(self, other, operator):
        """Takes another collection, and apply the operator to it, and returns a new instance.

        Operator must be a callable that accepts a set of the numbers of self,
        and another set for other's numbers.
        """
        if not isinstance(other, type(self)):
            raise TypeError(
                f"Other side must be of the type {type(self).__name__}. {other} given."
            )
        self_nums = set(self.keys())
        other_nums = set(other.keys())
        new_nums = operator(self_nums, other_nums)
        new_objs = {}
        # give preference to self
        for obj in it.chain(other, self):
            if obj.number in new_nums:
                new_objs[obj.number] = obj
        return type(self)(list(new_objs.values()))

    def __and__(self, other):
        return self.__set_logic(other, lambda a, b: a & b)

    def __iand__(self, other):
        new_vals = self & other
        self.__num_cache.clear()
        self._objects.clear()
        self.update(new_vals)
        return self

    def __or__(self, other):
        return self.__set_logic(other, lambda a, b: a | b)

    def __ior__(self, other):
        new_vals = other - self
        self.extend(new_vals)
        return self

    def __sub__(self, other):
        return self.__set_logic(other, lambda a, b: a - b)

    def __isub__(self, other):
        excess_values = self & other
        for excess in excess_values:
            del self[excess.number]
        return self

    def __xor__(self, other):
        return self.__set_logic(other, lambda a, b: a ^ b)

    def __ixor__(self, other):
        new_values = self ^ other
        self._objects.clear()
        self.__num_cache.clear()
        self.update(new_values)
        return self

    def __set_logic_test(self, other, operator):
        """Takes another collection, and apply the operator to it, testing the logic of it.

        Operator must be a callable that accepts a set of the numbers of self,
        and another set for other's numbers.
        """
        if not isinstance(other, type(self)):
            raise TypeError(
                f"Other side must be of the type {type(self).__name__}. {other} given."
            )
        self_nums = set(self.keys())
        other_nums = set(other.keys())
        return operator(self_nums, other_nums)

    def __le__(self, other):
        return self.__set_logic_test(other, lambda a, b: a <= b)

    def __lt__(self, other):
        return self.__set_logic_test(other, lambda a, b: a < b)

    def __ge__(self, other):
        return self.__set_logic_test(other, lambda a, b: a >= b)

    def __gt__(self, other):
        return self.__set_logic_test(other, lambda a, b: a > b)

    def issubset(self, other: typing.Self):
        """Test whether every element in the collection is in other.

        ``collection <= other``

        .. versionadded:: 1.0.0

        Parameters
        ----------
        other : Self
            the set to compare to.

        Returns
        -------
        bool
        """
        return self.__set_logic_test(other, lambda a, b: a.issubset(b))

    def isdisjoint(self, other: typing.Self):
        """Test if there are no elements in common between the collection, and other.

        Collections are disjoint if and only if their intersection
        is the empty set.

        .. versionadded:: 1.0.0

        Parameters
        ----------
        other : Self
            the set to compare to.

        Returns
        -------
        bool
        """
        return self.__set_logic_test(other, lambda a, b: a.isdisjoint(b))

    def issuperset(self, other: typing.Self):
        """Test whether every element in other is in the collection.

        ``collection >= other``

        .. versionadded:: 1.0.0

        Parameters
        ----------
        other : Self
            the set to compare to.

        Returns
        -------
        bool
        """
        return self.__set_logic_test(other, lambda a, b: a.issuperset(b))

    def __set_logic_multi(self, others, operator):
        for other in others:
            if not isinstance(other, type(self)):
                raise TypeError(
                    f"Other argument must be of type {type(self).__name__}. {other} given."
                )
        self_nums = set(self.keys())
        other_sets = []
        for other in others:
            other_sets.append(set(other.keys()))
        valid_nums = operator(self_nums, *other_sets)
        objs = {}
        for obj in it.chain(*others, self):
            if obj.number in valid_nums:
                objs[obj.number] = obj
        return type(self)(list(objs.values()))

    def intersection(self, *others: typing.Self):
        """Return a new collection with all elements in common in collection, and all others.

        ``collection & other & ...``

        .. versionadded:: 1.0.0

        Parameters
        ----------
        *others : Self
            the other collections to compare to.

        Returns
        -------
        typing.Self
        """
        return self.__set_logic_multi(others, lambda a, *b: a.intersection(*b))

    def intersection_update(self, *others: typing.Self):
        """Update the collection keeping all elements in common in collection, and all others.

        ``collection &= other & ...``

        .. versionadded:: 1.0.0

        Parameters
        ----------
        *others : Self
            the other collections to compare to.
        """
        if len(others) == 1:
            self &= others[0]
        else:
            other = others[0].intersection(*others[1:])
            self &= other

    def union(self, *others: typing.Self):
        """Return a new collection with all elements from collection, and all others.

        ``collection | other | ...``

        .. versionadded:: 1.0.0

        Parameters
        ----------
        *others : Self
            the other collections to compare to.

        Returns
        -------
        typing.Self
        """
        return self.__set_logic_multi(others, lambda a, *b: a.union(*b))

    def difference(self, *others: typing.Self):
        """Return a new collection with elements from collection, that are not in the others.

        ``collection - other - ...``

        .. versionadded:: 1.0.0

        Parameters
        ----------
        *others : Self
            the other collections to compare to.

        Returns
        -------
        typing.Self
        """
        return self.__set_logic_multi(others, lambda a, *b: a.difference(*b))

    def difference_update(self, *others: typing.Self):
        """Update the new collection removing all elements from others.

        ``collection -= other | ...``

        .. versionadded:: 1.0.0

        Parameters
        ----------
        *others : Self
            the other collections to compare to.
        """
        new_vals = self.difference(*others)
        self.clear()
        self.update(new_vals)
        return self

    def symmetric_difference(self, other: typing.Self):
        """Return a new collection with elements in either the collection or the other, but not both.

        ``collection ^ other``

        .. versionadded:: 1.0.0

        Parameters
        ----------
        others : Self
            the other collections to compare to.

        Returns
        -------
        typing.Self
        """
        return self ^ other

    def symmetric_difference_update(self, other: typing.Self):
        """Update the collection, keeping only elements found in either collection, but not in both.

        ``collection ^= other``

        .. versionadded:: 1.0.0

        Parameters
        ----------
        others : Self
            the other collections to compare to.
        """
        self ^= other
        return self

    def discard(self, obj: montepy.numbered_mcnp_object.Numbered_MCNP_Object):
        """Remove the object from the collection if it is present.

        .. versionadded:: 1.0.0

        Parameters
        ----------
        obj : Numbered_MCNP_Object
            the object to remove.
        """
        try:
            self.remove(obj)
        except (TypeError, KeyError) as e:
            pass

    def get(self, i: int, default=None) -> (Numbered_MCNP_Object, None):
        """Get ``i`` if possible, or else return ``default``.

        Parameters
        ----------
        i : int
            number of the object to get, not it's location in the
            internal list
        default : object
            value to return if not found

        Returns
        -------
        Numbered_MCNP_Object
        """
        try:
            ret = self.__num_cache[i]
            if ret.number == i:
                return ret
        except KeyError:
            pass
        for obj in self._objects:
            self.__num_cache[obj.number] = obj
            if obj.number == i:
                self.__num_cache[i] = obj
                return obj
        return default

    def keys(self) -> typing.Generator[int, None, None]:
        """Get iterator of the collection's numbers.

        Returns
        -------
        int
        """
        if len(self) == 0:
            yield from []
        for o in self._objects:
            self.__num_cache[o.number] = o
            yield o.number

    def values(self) -> typing.Generator[Numbered_MCNP_Object, None, None]:
        """Get iterator of the collection's objects.

        Returns
        -------
        Numbered_MCNP_Object
        """
        for o in self._objects:
            self.__num_cache[o.number] = o
            yield o

    def items(
        self,
    ) -> typing.Generator[typing.Tuple[int, Numbered_MCNP_Object], None, None]:
        """Get iterator of the collections (number, object) pairs.

        Returns
        -------
        tuple(int, MCNP_Object)
        """
        for o in self._objects:
            yield o.number, o

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            raise TypeError(
                f"Can only compare {type(self).__name__} to each other. {other} was given."
            )
        if len(self) != len(other):
            return False
        keys = sorted(self.keys())
        for key in keys:
            try:
                if self[key] != other[key]:
                    return False
            except KeyError:
                return False
        return True


class NumberedDataObjectCollection(NumberedObjectCollection):
    def __init__(self, obj_class, objects=None, problem=None):
        self._last_index = None
        if problem and objects:
            try:
                self._last_index = problem.data_inputs.index(objects[-1])
            except ValueError:
                pass
        super().__init__(obj_class, objects, problem)

    def _append_hook(self, obj, insert_in_data=True):
        """Appends the given object to the end of this collection.

        Parameters
        ----------
        obj : Numbered_MCNP_Object
            the object to add.
        insert_in_data : bool
            Whether to add the object to the linked problem's
            data_inputs.

        Raises
        ------
        NumberConflictError
            if this object has a number that is already in use.
        """
        if self._problem:
            if self._last_index:
                index = self._last_index
            elif len(self) > 0:
                try:
                    index = self._problem.data_inputs.index(self._objects[-1])
                except ValueError:
                    index = len(self._problem.data_inputs)
            else:
                index = len(self._problem.data_inputs)
            if insert_in_data:
                self._problem.data_inputs.insert(index + 1, obj)
            self._last_index = index + 1

    def _delete_hook(self, obj):
        if self._problem:
            self._problem.data_inputs.remove(obj)

    def clear(self):
        """Removes all objects from this collection."""
        if self._problem:
            for obj in self._objects:
                self._problem.data_inputs.remove(obj)
        self._last_index = None
        super().clear()



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from enum import Enum


class Shortcuts(Enum):
    """Enumeration of the possible MCNP shortcuts."""

    REPEAT = "r"
    """A repeated entry shortcut."""
    JUMP = "j"
    """A jump entry, which counts as a default entry."""
    INTERPOLATE = "i"
    """A linear interpolation."""
    LOG_INTERPOLATE = "ilog"
    """A logarithmic interpolation."""
    MULTIPLY = "m"
    """a multiplication of the previous entry."""



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.


class MaterialComponent:
    """A class to represent a single component in a material.

    For example: this may be H-1 in water: like 1001.80c — 0.6667

    .. deprecated:: 0.4.1
        MaterialComponent has been deprecated as part of a redesign for the material
        interface due to a critical bug in how MontePy handles duplicate nuclides.
        It has been removed in 1.0.0.
        See :ref:`migrate 0 1`.

    Raises
    ------
    DeprecationWarning
        whenever called.
    """

    def __init__(self, *args):
        raise DeprecationWarning(
            f"""MaterialComponent is deprecated, and has been removed in MontePy 1.0.0.
See <https://www.montepy.org/migrations/migrate0_1.html> for more information """,
        )



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from abc import ABC, abstractmethod
import math
from montepy.errors import *
from montepy.input_parser.block_type import BlockType
from montepy.constants import BLANK_SPACE_CONTINUE, get_max_line_length
from montepy.input_parser.read_parser import ReadParser
from montepy.input_parser.tokens import CellLexer, SurfaceLexer, DataLexer
from montepy.utilities import *
import re


class Jump:
    """Class to represent a default entry represented by a "jump".

    |     I get up and nothing gets me down
    |     You got it tough, I've seen the toughest around
    |     And I know, baby, just how you feel
    |     You gotta roll with the punches to get to what's real

    |     Oh, can't you see me standing here?
    |     I got my back against the record machine
    |     I ain't the worst that you've seen
    |     Oh, can't you see what I mean?

    |    Ah, might as well ...
    """

    def __str__(self):
        return "J"

    def __repr__(self):
        return f"Jump: {hex(id(self))}"

    def __format__(self, spec):
        return format(str(self), spec)

    def __bool__(self):
        raise TypeError("Jump doesn't have a truthiness or falsiness")

    def __eq__(self, other):
        return type(self) == type(other)

    def lower(self):
        """Hop.

        Returns
        -------
        str
        """
        return "j"

    def title(self):
        """Skip.

        Returns
        -------
        str
        """
        return "Jump"

    def upper(self):
        """Jump.

        Returns
        -------
        str
        """
        return "J"


class ParsingNode(ABC):
    """Object to represent a single coherent MCNP input, such as an input.

    Parameters
    ----------
    input_lines : list
        the lines read straight from the input file.
    """

    def __init__(self, input_lines):
        if not isinstance(input_lines, list):
            raise TypeError("input_lines must be a list")
        for line in input_lines:
            if not isinstance(line, str):
                raise TypeError(f"element: {line} in input_lines must be a string")
        self._input_lines = input_lines

    @property
    def input_lines(self):
        """The lines of the input read straight from the input file

        Returns
        -------
        list
        """
        return self._input_lines

    @property
    def input_text(self):
        return "\n".join(self.input_lines) + "\n"

    @abstractmethod
    def format_for_mcnp_input(self, mcnp_version):
        """Creates a string representation of this input that can be
        written to file.

        Parameters
        ----------
        mcnp_version : tuple
            The tuple for the MCNP version that must be exported to.

        Returns
        -------
        list
            a list of strings for the lines that this input will occupy.
        """
        pass


class Input(ParsingNode):
    """Represents a single MCNP "Input" e.g. a single cell definition.

    Parameters
    ----------
    input_lines : list
        the lines read straight from the input file.
    block_type : BlockType
        An enum showing which of three MCNP blocks this was inside of.
    input_file : MCNP_InputFile
        the wrapper for the input file this is read from.
    lineno : int
        the line number this input started at. 1-indexed.
    """

    SPECIAL_COMMENT_PREFIXES = ["fc", "sc"]
    """Prefixes for special comments like tally comments.

    Returns
    -------
    list
    """

    def __init__(self, input_lines, block_type, input_file=None, lineno=None):
        super().__init__(input_lines)
        if not isinstance(block_type, BlockType):
            raise TypeError("block_type must be BlockType")
        self._block_type = block_type
        self._input_file = input_file
        self._lineno = lineno
        self._lexer = None

    def __str__(self):
        return f"INPUT: {self._block_type}"

    def __repr__(self):
        return f"INPUT: {self._block_type}: {self.input_lines}"

    @property
    def block_type(self):
        """Enum representing which block of the MCNP input this came from.

        Returns
        -------
        BlockType
        """
        return self._block_type

    @make_prop_pointer("_input_file")
    def input_file(self):
        """The file this input file was read from.

        Returns
        -------
        MCNP_InputFile
        """
        pass

    @make_prop_pointer("_lineno")
    def line_number(self):
        """The line number this input started on.

        This is 1-indexed.

        Returns
        -------
        int
        """
        pass

    def format_for_mcnp_input(self, mcnp_version):
        pass

    def tokenize(self):
        """Tokenizes this input as a stream of Tokens.

        This is a generator of Tokens.
        This is context dependent based on :func:`block_type`.

        * In a cell block :class:`~montepy.input_parser.tokens.CellLexer` is used.
        * In a surface block :class:`~montepy.input_parser.tokens.SurfaceLexer` is used.
        * In a data block :class:`~montepy.input_parser.tokens.DataLexer` is used.

        Returns
        -------
        Token
            a generator of tokens.
        """
        if self.block_type == BlockType.CELL:
            lexer = CellLexer()
        elif self.block_type == BlockType.SURFACE:
            lexer = SurfaceLexer()
        else:
            lexer = DataLexer()
        self._lexer = lexer
        # hacky way to capture final new line and remove it after lexing.
        generator = lexer.tokenize(self.input_text)
        token = None
        next_token = None
        try:
            token = next(generator)
            while True:
                if next_token:
                    token = next_token
                next_token = next(generator)
                yield token
        except StopIteration:
            if not token:
                return
            token.value = token.value.rstrip("\n")
            if token.value:
                yield token
        self._lexer = None

    @make_prop_pointer("_lexer")
    def lexer(self):
        """The current lexer being used to parse this input.

        If not currently tokenizing this will be None.

        Returns
        -------
        MCNP_Lexer
        """
        pass


class ReadInput(Input):
    """A input for the read input that reads another input file

    Parameters
    ----------
    input_lines : list
        the lines read straight from the input file.
    block_type : BlockType
        An enum showing which of three MCNP blocks this was inside of.
    input_file : MCNP_InputFile
        the wrapper for the input file this is read from.
    lineno : int
        the line number this input started at. 1-indexed.
    """

    _parser = ReadParser()

    def __init__(self, input_lines, block_type, input_file=None, lineno=None):
        super().__init__(input_lines, block_type, input_file, lineno)
        if not self.is_read_input(input_lines):
            raise ValueError("Not a valid Read Input")
        parse_result = self._parser.parse(self.tokenize(), self)
        if not parse_result:
            raise ParsingError(self, "", self._parser.log.clear_queue())
        self._tree = parse_result
        self._parameters = self._tree["parameters"]

    @staticmethod
    def is_read_input(input_lines):
        first_non_comment = ""
        for line in input_lines:
            if not is_comment(line):
                first_non_comment = line
                break
        words = first_non_comment.split()
        if len(words) > 0:
            first_word = words[0].lower()
            return first_word == "read"
        # this is a fall through catch that only happens for a blank input
        return False  # pragma: no cover

    @property
    def file_name(self):
        """The relative path to the filename specified in this read input.

        Returns
        -------
        str
        """
        return self._parameters["file"]["data"].value

    def __str__(self):
        return f"READ INPUT: Block_Type: {self.block_type}"

    def __repr__(self):
        return (
            f"READ INPUT: {self._block_type}: {self.input_lines} File: {self.file_name}"
        )


class Message(ParsingNode):
    """Object to represent an MCNP message.

    These are blocks at the beginning of an input that are printed in the output.

    Parameters
    ----------
    input_lines : list
        the lines read straight from the input file.
    lines : list
        the strings of each line in the message block
    """

    def __init__(self, input_lines, lines):
        super().__init__(input_lines)
        if not isinstance(lines, list):
            raise TypeError("lines must be a list")
        for line in lines:
            if not isinstance(line, str):
                raise TypeError(f"line {line} in lines must be a string")
        buff = []
        for line in lines:
            buff.append(line.rstrip())
        self._lines = buff

    def __str__(self):
        return f"MESSAGE: {len(self._lines)} lines"

    def __repr__(self):
        ret = "MESSAGE:\n"
        for line in self._lines:
            ret += line + "\n"
        return ret

    @property
    def lines(self):
        """The lines of input for the message block.

        Each entry is a string of that line in the message block

        Returns
        -------
        list
        """
        return self._lines

    def format_for_mcnp_input(self, mcnp_version):
        ret = []
        line_length = get_max_line_length(mcnp_version)
        for i, line in enumerate(self.lines):
            if i == 0:
                ret.append("MESSAGE: " + line[0 : line_length - 10])
            else:
                ret.append(line[0 : line_length - 1])
        ret.append("")
        return ret


class Title(ParsingNode):
    """Object to represent the title for an MCNP problem

    Parameters
    ----------
    input_lines : list
        the lines read straight from the input file.
    title : str
        The string for the title of the problem.
    """

    def __init__(self, input_lines, title):
        """
        Parameters
        ----------
        input_lines : list
            the lines read straight from the input file.
        title : str
            The string for the title of the problem.
        """
        super().__init__(input_lines)
        if not isinstance(title, str):
            raise TypeError("title must be a string")
        self._title = title.rstrip()

    @property
    def title(self):
        """The string of the title set for this problem

        Returns
        -------
        str
        """
        return self._title

    def __str__(self):
        return f"TITLE: {self._title}"

    def format_for_mcnp_input(self, mcnp_version):
        line_length = 0
        line_length = get_max_line_length(mcnp_version)
        return [self.title[0 : line_length - 1]]



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
import pytest

import montepy
from montepy.mcnp_problem import MCNP_Problem


@pytest.fixture(scope="module")
def problem_path():
    return "tests/inputs/test.imcnp"


@pytest.fixture(scope="module")
def problem(problem_path):
    return MCNP_Problem(problem_path)


def test_problem_init(problem, problem_path):
    assert isinstance(
        problem.input_file, montepy.input_parser.input_file.MCNP_InputFile
    )
    assert problem.input_file.path == problem_path
    assert problem.input_file.name == problem_path
    assert problem.mcnp_version == (6, 3, 0)


def test_problem_str(problem, problem_path):
    assert f"MCNP problem for: {problem_path}" in str(problem)


def test_problem_repr(problem, problem_path):
    assert repr(problem).startswith(f"MCNP problem for: {problem_path}")


def test_problem_write_type(problem):
    unwritable = object()
    with pytest.raises(TypeError):
        problem.write_problem(unwritable)


def test_problem_clone(problem):
    problem.parse_input()
    new_problem = problem.clone()
    for collection in {"cells", "surfaces", "materials"}:
        for old_obj, new_obj in zip(
            getattr(problem, collection), getattr(new_problem, collection)
        ):
            assert old_obj is not new_obj
            assert new_obj._problem is new_problem



# Copyright 2024-2025, Battelle Energy Alliance, LLC All Rights Reserved.
# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
"""MontePy is a library for reading, editing, and writing MCNP input files.

This creates a semantic understanding of the MCNP input file.
start by running montepy.read_input().

You will receive an MCNP_Problem object that you will interact with.
"""

from . import data_inputs
from . import input_parser
from . import constants
import importlib.metadata

# data input promotion

from montepy.data_inputs.material import Material
from montepy.data_inputs.transform import Transform
from montepy.data_inputs.nuclide import Library, Nuclide
from montepy.data_inputs.element import Element
from montepy.data_inputs.thermal_scattering import ThermalScatteringLaw
from montepy.data_inputs.data_parser import parse_data

# geometry
from montepy.geometry_operators import Operator
from montepy import geometry_operators
from montepy.surfaces.surface_type import SurfaceType
from montepy.surfaces import *

# input parser
from montepy.input_parser.mcnp_input import Jump
from .input_parser.input_reader import read_input

# top level
from montepy.particle import Particle, LibraryType
from montepy.universe import Universe
from montepy.cell import Cell
from montepy.mcnp_problem import MCNP_Problem

# collections
from montepy.cells import Cells
from montepy.materials import Materials
from montepy.universes import Universes
from montepy.surface_collection import Surfaces
from montepy.transforms import Transforms

import montepy.errors
import sys


try:
    from . import _version

    __version__ = _version.version
except ImportError:
    try:
        from setuptools_scm import get_version

        __version__ = get_version()
    except (ImportError, LookupError):
        __version__ = "Undefined"


# enable deprecated warnings for users
if not sys.warnoptions:
    import os, warnings

    warnings.simplefilter("default")  # Change the filter in this process
    os.environ["PYTHONWARNINGS"] = "default"  # Also affect subprocesses
